# -*- coding: utf-8 -*-
#
# Copyright 2017-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Print a collection as a table."""

from datetime import datetime
from operator import attrgetter

from tabulate import tabulate as tblte


def format_cell(cell, datetime_fmt=None):
    """Format a cell."""
    if datetime_fmt and isinstance(cell, datetime):
        if cell.tzinfo:
            cell = cell.astimezone()
        return cell.strftime(datetime_fmt)
    return cell


def tabulate(collection, headers, datetime_fmt='%Y-%m-%d %H:%M:%S', **kwargs):
    """Pretty-print a collection."""
    if isinstance(headers, dict):
        attrs = headers.keys()
        # if mapping is not specified keep original
        names = [
            key if value is None else value for key, value in headers.items()
        ]
    else:
        attrs = names = headers
    table = [(
        format_cell(cell, datetime_fmt=datetime_fmt)
        for cell in _to_list(attrgetter(*attrs)(c))
    ) for c in collection]
    return tblte(table, headers=[h.upper() for h in names], **kwargs)


def _to_list(value):
    if isinstance(value, (list, tuple)):
        return value
    return [value]



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``dataset`` command."""

from __future__ import absolute_import, print_function

import json
import os
from pathlib import Path

import pytest
import requests
import yaml

from renku.cli import cli
from renku.core.commands.format.dataset_files import DATASET_FILES_COLUMNS, \
    DATASET_FILES_FORMATS
from renku.core.commands.format.datasets import DATASETS_COLUMNS, \
    DATASETS_FORMATS
from renku.core.commands.providers import DataverseProvider, ProviderFactory, \
    ZenodoProvider
from renku.core.management.config import RENKU_HOME
from renku.core.management.datasets import DatasetsApiMixin
from renku.core.models.refs import LinkReference
from renku.core.utils.contexts import chdir
from renku.core.utils.datetime8601 import validate_iso8601


def test_datasets_create_clean(runner, project, client):
    """Test creating a dataset in clean repository."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    dataset = client.load_dataset('dataset')
    assert dataset

    staged = client.repo.index.diff('HEAD')
    for file_path in staged:
        assert 'datasets' not in file_path

    untracked = client.repo.untracked_files
    for file_path in untracked:
        assert 'datasets' not in file_path


def test_datasets_create_with_metadata(runner, client):
    """Test creating a dataset with metadata."""
    result = runner.invoke(
        cli, [
            'dataset', 'create', 'my-dataset', '--title', 'Long Title',
            '--description', 'some description here', '-c',
            'John Doe <john.doe@mail.ch>', '-c',
            'John Smiths<john.smiths@mail.ch>'
        ]
    )
    assert 0 == result.exit_code
    assert 'OK' in result.output

    dataset = client.load_dataset('my-dataset')
    assert dataset.name == 'Long Title'
    assert dataset.short_name == 'my-dataset'
    assert dataset.description == 'some description here'
    assert 'John Doe' in [c.name for c in dataset.creator]
    assert 'john.doe@mail.ch' in [c.email for c in dataset.creator]
    assert 'John Smiths' in [c.name for c in dataset.creator]
    assert 'john.smiths@mail.ch' in [c.email for c in dataset.creator]


def test_datasets_create_different_names(runner, client):
    """Test creating datasets with same title but different short_name."""
    result = runner.invoke(
        cli, ['dataset', 'create', 'dataset-1', '--title', 'title']
    )
    assert 0 == result.exit_code
    assert 'OK' in result.output

    result = runner.invoke(
        cli, ['dataset', 'create', 'dataset-2', '--title', 'title']
    )
    assert 0 == result.exit_code
    assert 'OK' in result.output


def test_datasets_create_with_same_name(runner, client):
    """Test creating datasets with same name."""
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 1 == result.exit_code
    assert 'Dataset exists: "dataset"' in result.output


@pytest.mark.parametrize(
    'name', [
        'any name /@#$!', 'name longer than 24 characters', 'semi valid-name',
        'dataset/new', '/dataset', 'dataset/'
    ]
)
def test_datasets_invalid_name(runner, client, name):
    """Test creating datasets with invalid name."""
    result = runner.invoke(cli, ['dataset', 'create', name])
    assert 2 == result.exit_code
    assert 'short_name "{}" is not valid.'.format(name) in result.output


def test_datasets_create_dirty(runner, project, client):
    """Test creating a dataset in dirty repository."""
    # Create a file in root of the repository.
    with (client.path / 'a').open('w') as fp:
        fp.write('a')

    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    dataset = client.load_dataset('dataset')
    assert dataset

    staged = client.repo.index.diff('HEAD')
    for file_path in staged:
        assert 'datasets' not in file_path

    untracked = client.repo.untracked_files
    for file_path in untracked:
        assert 'datasets' not in file_path


def test_datasets_create_dirty_exception_untracked(runner, project, client):
    """Test exception raise for untracked file in renku directory."""
    # 1. Create a problem.
    datasets_dir = client.path / RENKU_HOME / DatasetsApiMixin.DATASETS
    if not datasets_dir.exists():
        datasets_dir.mkdir()

    with (datasets_dir / 'a').open('w') as fp:
        fp.write('a')

    # 2. Ensure correct error has been raised.
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 1 == result.exit_code
    assert '.renku contains uncommitted changes.' in result.output


def test_datasets_create_dirty_exception_staged(runner, project, client):
    """Test exception raise for staged file in renku directory."""
    # 1. Create a problem within .renku directory
    datasets_dir = client.path / RENKU_HOME / DatasetsApiMixin.DATASETS
    if not datasets_dir.exists():
        datasets_dir.mkdir()

    with (datasets_dir / 'a').open('w') as fp:
        fp.write('a')

    # 2. Stage a problem without committing it.
    client.repo.git.add(datasets_dir / 'a')

    # 3. Ensure correct error has been raised.
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 1 == result.exit_code
    assert '.renku contains uncommitted changes.' in result.output


def test_dataset_create_dirty_exception_all_untracked(runner, project, client):
    """Test exception raise for all untracked files."""
    # 1. Create unclean root to enforce ensure checks.
    with (client.path / 'a').open('w') as fp:
        fp.write('a')

    # 2. Create a problem.
    datasets_dir = client.path / RENKU_HOME / DatasetsApiMixin.DATASETS
    if not datasets_dir.exists():
        datasets_dir.mkdir()

    with (datasets_dir / 'a').open('w') as fp:
        fp.write('a')

    # 3. Ensure correct error has been raised.
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 1 == result.exit_code
    assert '.renku contains uncommitted changes.' in result.output


def test_datasets_create_dirty_exception_all_staged(runner, project, client):
    """Test exception raise for all staged files."""
    # 1. Create unclean root to enforce ensure checks.
    with (client.path / 'a').open('w') as fp:
        fp.write('a')

    client.repo.git.add('a')

    # 2. Create a problem.
    datasets_dir = client.path / RENKU_HOME / DatasetsApiMixin.DATASETS
    if not datasets_dir.exists():
        datasets_dir.mkdir()

    with (datasets_dir / 'a').open('w') as fp:
        fp.write('a')

    client.repo.git.add(datasets_dir / 'a')

    # 3. Ensure correct error has been raised.
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 1 == result.exit_code
    assert '.renku contains uncommitted changes.' in result.output


def test_dataset_create_exception_refs(runner, project, client):
    """Test untracked/unstaged exception raise in dirty renku home dir."""
    with (client.path / 'a').open('w') as fp:
        fp.write('a')

    datasets_dir = client.path / RENKU_HOME / DatasetsApiMixin.DATASETS
    if not datasets_dir.exists():
        datasets_dir.mkdir()

    with (datasets_dir / 'a').open('w') as fp:
        fp.write('a')

    refs_dir = client.path / RENKU_HOME / LinkReference.REFS
    if not refs_dir.exists():
        refs_dir.mkdir()

    with (refs_dir / 'b').open('w') as fp:
        fp.write('b')

    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 1 == result.exit_code
    assert 'a' in result.output


@pytest.mark.parametrize(
    'creator,field', [('John Doe', 'Email'), ('John Doe<>', 'Email'),
                      ('<john.doe@mail.ch>', 'Name'),
                      ('John Doe<john.doe@mail>', 'Email')]
)
def test_dataset_creator_is_invalid(client, runner, creator, field):
    """Test create dataset with invalid creator format."""
    result = runner.invoke(cli, ['dataset', 'create', 'ds', '-c', creator])
    assert 2 == result.exit_code
    assert field + ' is invalid' in result.output


@pytest.mark.parametrize('output_format', DATASETS_FORMATS.keys())
def test_datasets_list_empty(output_format, runner, project):
    """Test listing without datasets."""
    format_option = '--format={0}'.format(output_format)
    result = runner.invoke(cli, ['dataset', format_option])
    assert 0 == result.exit_code


@pytest.mark.parametrize('output_format', DATASETS_FORMATS.keys())
def test_datasets_list_non_empty(output_format, runner, project):
    """Test listing with datasets."""
    format_option = '--format={0}'.format(output_format)
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    result = runner.invoke(cli, ['dataset', format_option])
    assert 0 == result.exit_code
    assert 'my-dataset' in result.output

    result = runner.invoke(
        cli, ['dataset', '--revision=HEAD~1', format_option]
    )
    assert 0 == result.exit_code
    assert 'my-dataset' not in result.output


@pytest.mark.parametrize(
    'columns,headers,values', [(
        'title,short_name', ['TITLE', 'SHORT_NAME'
                             ], ['my-dataset', 'Long Title']
    ), ('creators', ['CREATORS'], ['John Doe'])]
)
def test_datasets_list_with_columns(runner, project, columns, headers, values):
    """Test listing datasets with custom column name."""
    result = runner.invoke(
        cli, [
            'dataset', 'create', 'my-dataset', '--title', 'Long Title', '-c',
            'John Doe <john.doe@mail.ch>'
        ]
    )
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['dataset', '--columns', columns])
    assert 0 == result.exit_code
    assert headers == result.output.split('\n').pop(0).split()
    for value in values:
        assert value in result.output


@pytest.mark.parametrize('column', DATASETS_COLUMNS.keys())
def test_datasets_list_columns_correctly(runner, project, column):
    """Test dataset listing only shows requested columns."""
    result = runner.invoke(cli, ['dataset', '--columns', column])
    assert 0 == result.exit_code
    header = result.output.split('\n').pop(0)
    name, display_name = DATASETS_COLUMNS[column]
    display_name = display_name or name
    assert display_name.upper() == header


@pytest.mark.parametrize('columns', ['invalid', 'id,invalid'])
def test_datasets_list_invalid_column(runner, project, columns):
    """Test dataset listing invalid column name."""
    result = runner.invoke(cli, ['dataset', '--columns', columns])
    assert 2 == result.exit_code
    assert 'Invalid column name: "invalid".' in result.output


def test_add_and_create_dataset(directory_tree, runner, project, client):
    """Test add data to a non-existing dataset."""
    result = runner.invoke(
        cli, ['dataset', 'add', 'new-dataset',
              str(directory_tree)],
        catch_exceptions=False
    )
    assert 1 == result.exit_code
    assert 'Dataset "new-dataset" does not exist.' in result.output

    # Add succeeds with --create
    result = runner.invoke(
        cli,
        ['dataset', 'add', '--create', 'new-dataset',
         str(directory_tree)],
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    # Further, add with --create fails
    result = runner.invoke(
        cli,
        ['dataset', 'add', '--create', 'new-dataset',
         str(directory_tree)],
        catch_exceptions=False
    )
    assert 1 == result.exit_code


def test_add_to_dirty_repo(directory_tree, runner, project, client):
    """Test adding to a dataset in a dirty repo commits only added files."""
    with (client.path / 'tracked').open('w') as fp:
        fp.write('tracked file')
    client.repo.git.add('*')
    client.repo.index.commit('tracked file')

    with (client.path / 'tracked').open('w') as fp:
        fp.write('modified tracked file')
    with (client.path / 'untracked').open('w') as fp:
        fp.write('untracked file')

    result = runner.invoke(
        cli,
        ['dataset', 'add', '--create', 'new-dataset',
         str(directory_tree)],
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    assert client.repo.is_dirty()
    assert ['untracked'] == client.repo.untracked_files

    # Add without making a change
    result = runner.invoke(
        cli, ['dataset', 'add', 'new-dataset',
              str(directory_tree)],
        catch_exceptions=False
    )
    assert 1 == result.exit_code
    assert 'Error: File already exists in dataset' in result.output

    assert client.repo.is_dirty()
    assert ['untracked'] == client.repo.untracked_files


def test_multiple_file_to_dataset(tmpdir, runner, project, client):
    """Test importing multiple data into a dataset at once."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    with client.with_dataset('dataset') as dataset:
        assert dataset.name == 'dataset'

    paths = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))

    # add data
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'dataset'] + paths,
        catch_exceptions=False,
    )
    assert 0 == result.exit_code


def test_repository_file_to_dataset(runner, project, client):
    """Test adding a file from the repository into a dataset."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    with (client.path / 'a').open('w') as fp:
        fp.write('a')

    client.repo.git.add('a')
    client.repo.git.commit(message='Added file a')

    # add data
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'dataset', 'a'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    with client.with_dataset('dataset') as dataset:
        assert dataset.name == 'dataset'
        assert dataset.find_file('a') is not None


def test_relative_import_to_dataset(tmpdir, runner, client):
    """Test importing data from a directory structure."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    with client.with_dataset('dataset') as dataset:
        assert dataset.name == 'dataset'

    zero_data = tmpdir.join('zero.txt')
    zero_data.write('zero')

    first_level = tmpdir.mkdir('first')
    second_level = first_level.mkdir('second')

    first_data = first_level.join('first.txt')
    first_data.write('first')

    second_data = second_level.join('second.txt')
    second_data.write('second')

    paths = [str(zero_data), str(first_level), str(second_level)]

    # add data in subdirectory
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'dataset'] + paths,
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    assert os.stat(os.path.join('data', 'dataset', 'zero.txt'))
    assert os.stat(os.path.join('data', 'dataset', 'first', 'first.txt'))
    assert os.stat(
        os.path.join('data', 'dataset', 'first', 'second', 'second.txt')
    )


@pytest.mark.parametrize(
    'params,message', [
        (['-s', 'file', 'https://example.com'
          ], 'Cannot use "--source" with URLs or local files.'),
        (['-s', 'file', '/some/local/path'
          ], 'Cannot use "--source" with URLs or local files.'),
    ]
)
def test_usage_error_in_add_from_url(runner, client, params, message):
    """Test user's errors when adding URL/local file to a dataset."""
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'remote', '--create'] + params,
        catch_exceptions=False,
    )
    assert 2 == result.exit_code
    assert message in result.output


def test_add_from_local_repo_warning(
    runner, client, data_repository, directory_tree
):
    """Test a warning is printed when adding from a local git repo."""
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'dataset', '--create',
         str(directory_tree)],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code
    assert 'Use remote\'s Git URL instead to enable lineage ' in result.output


def test_dataset_add_with_link(tmpdir, runner, project, client):
    """Test adding data to dataset with --link flag."""
    import stat

    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    paths = []
    expected_inodes = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        expected_inodes.append(os.lstat(str(new_file))[stat.ST_INO])
        paths.append(str(new_file))

    # add data
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset', '--link'] + paths,
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    received_inodes = []
    with client.with_dataset('my-dataset') as dataset:
        assert dataset.name == 'my-dataset'
        for file_ in dataset.files:
            path_ = (client.path / file_.path).resolve()
            received_inodes.append(os.lstat(str(path_))[stat.ST_INO])

    # check that original and dataset inodes are the same
    assert sorted(expected_inodes) == sorted(received_inodes)


def test_dataset_add_with_copy(tmpdir, runner, project, client):
    """Test adding data to dataset with copy."""
    import os
    import stat

    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    paths = []
    original_inodes = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        original_inodes.append(os.lstat(str(new_file))[stat.ST_INO])
        paths.append(str(new_file))

    # add data
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset'] + paths,
    )
    assert 0 == result.exit_code

    received_inodes = []
    with client.with_dataset('my-dataset') as dataset:
        assert dataset.name == 'my-dataset'

        for file_ in dataset.files:
            path_ = (client.path / file_.path).resolve()
            received_inodes.append(os.lstat(str(path_))[stat.ST_INO])

    # check that original inodes are within created ones
    for inode in received_inodes:
        assert inode not in original_inodes


def test_dataset_file_path_from_subdirectory(runner, project, client):
    """Test adding a file into a dataset and check path independent
    of the CWD """
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    with (client.path / 'a').open('w') as fp:
        fp.write('a')

    client.repo.git.add('a')
    client.repo.git.commit(message='Added file a')

    # add data
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'dataset', 'a'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    with client.with_dataset('dataset') as dataset:
        datasetfile = dataset.find_file('a')
        assert datasetfile is not None

        assert datasetfile.full_path == client.path / 'a'

        os.chdir('./data')

        assert datasetfile.full_path == client.path / 'a'


def test_datasets_ls_files_tabular_empty(runner, project):
    """Test listing of data within empty dataset."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # list all files in dataset
    result = runner.invoke(
        cli, [
            'dataset', 'ls-files', '--columns', 'added,creators,dataset,path',
            'my-dataset'
        ]
    )
    assert 0 == result.exit_code

    # check output
    output = result.output.split('\n')
    assert output.pop(0).split() == ['ADDED', 'CREATORS', 'DATASET', 'PATH']
    assert set(output.pop(0)) == {' ', '-'}
    assert output.pop(0) == ''
    assert not output


@pytest.mark.parametrize('output_format', DATASET_FILES_FORMATS.keys())
def test_datasets_ls_files_check_exit_code(output_format, runner, project):
    """Test file listing exit codes for different formats."""
    format_option = '--format={0}'.format(output_format)
    result = runner.invoke(cli, ['dataset', 'ls-files', format_option])
    assert 0 == result.exit_code


@pytest.mark.parametrize('column', DATASET_FILES_COLUMNS.keys())
def test_datasets_ls_files_columns_correctly(runner, project, column):
    """Test file listing only shows requested columns."""
    result = runner.invoke(cli, ['dataset', 'ls-files', '--columns', column])
    assert 0 == result.exit_code
    header = result.output.split('\n').pop(0)
    name, display_name = DATASET_FILES_COLUMNS[column]
    display_name = display_name or name
    assert display_name.upper() == header


@pytest.mark.parametrize('columns', ['invalid', 'path,invalid'])
def test_datasets_ls_files_invalid_column(runner, project, columns):
    """Test file listing with invalid column name."""
    result = runner.invoke(cli, ['dataset', 'ls-files', '--columns', columns])
    assert 2 == result.exit_code
    assert 'Invalid column name: "invalid".' in result.output


def test_datasets_ls_files_tabular_dataset_filter(tmpdir, runner, project):
    """Test listing of data within dataset."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    paths = []
    created_files = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))
        created_files.append(new_file.basename)

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset'] + paths,
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    # list all files in non empty dataset
    result = runner.invoke(
        cli, ['dataset', 'ls-files', '--columns', 'added,path', 'my-dataset']
    )
    assert 0 == result.exit_code

    # check output from ls-files command
    output = result.output.split('\n')
    assert output.pop(0).split() == ['ADDED', 'PATH']
    assert set(output.pop(0)) == {' ', '-'}

    # check listing
    added_at = []
    for i in range(3):
        row = output.pop(0).split(' ')
        assert Path(row.pop()).name in created_files
        added_at.append(row.pop(0))

    # check if sorted by added_at
    assert added_at == sorted(added_at)


def test_datasets_ls_files_tabular_patterns(tmpdir, runner, project):
    """Test listing of data within dataset with include/exclude filters."""

    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    subdir = tmpdir.mkdir('sub')
    paths = [str(subdir)]
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))

        sub_file = subdir.join('sub_file_{0}'.format(i))
        sub_file.write(str(i))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset'] + paths,
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    # check include / exclude filters
    result = runner.invoke(
        cli,
        ['dataset', 'ls-files', '--include=**/file*', '--exclude=**/file_2']
    )
    assert 0 == result.exit_code

    # check output
    assert 'file_0' in result.output
    assert 'file_1' in result.output
    assert 'file_2' not in result.output

    # check directory pattern
    result = runner.invoke(cli, ['dataset', 'ls-files', '--include=**/sub/*'])
    assert 0 == result.exit_code

    # check output
    assert 'sub_file_0' in result.output
    assert 'sub_file_1' in result.output
    assert 'sub_file_2' in result.output


def test_datasets_ls_files_tabular_creators(tmpdir, runner, project, client):
    """Test listing of data within dataset with creators filters."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    paths = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset'] + paths,
    )
    assert 0 == result.exit_code

    creator = None
    with client.with_dataset('my-dataset') as dataset:
        creator = dataset.creator[0].name

    assert creator is not None
    assert len(dataset.creator) > 0

    # check creators filters
    result = runner.invoke(
        cli, ['dataset', 'ls-files', '--creators={0}'.format(creator)]
    )
    assert 0 == result.exit_code

    # check output
    for file_ in paths:
        assert str(Path(file_).name) in result.output


def test_datasets_ls_files_correct_paths(tmpdir, runner, project):
    """Test listing of data within dataset and check that paths are correct."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    paths = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset'] + paths,
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    # check include / exclude filters
    result = runner.invoke(cli, ['dataset', 'ls-files', '--format=json-ld'])
    assert 0 == result.exit_code

    output = json.loads(result.output)
    for record in output:
        assert Path(record['path']).exists()


def test_datasets_ls_files_with_name(directory_tree, runner, project):
    """Test listing of data within dataset with include/exclude filters."""
    # create a dataset
    result = runner.invoke(
        cli, ['dataset', 'create', 'my-dataset', '--title', 'Long Title']
    )
    assert 0 == result.exit_code

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset', directory_tree.strpath],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    # list files with short_name
    result = runner.invoke(cli, ['dataset', 'ls-files', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'dir2/file2' in result.output


def test_dataset_unlink_file_not_found(runner, project):
    """Test unlinking of file from dataset with no files found."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    result = runner.invoke(
        cli, ['dataset', 'unlink', 'my-dataset', '--include', 'notthere.csv']
    )

    assert 2 == result.exit_code


def test_dataset_unlink_file_abort_unlinking(tmpdir, runner, project):
    """Test unlinking of file from dataset and aborting."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    # unlink file from dataset
    result = runner.invoke(
        cli,
        ['dataset', 'unlink', 'my-dataset', '--include', new_file.basename],
        input='n'
    )
    assert 1 == result.exit_code

    # check output
    assert 'Aborted!' in result.output


def test_dataset_unlink_file(tmpdir, runner, client):
    """Test unlinking of file and check removal from dataset"""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    with client.with_dataset('my-dataset') as dataset:
        assert new_file.basename in {
            Path(file_.path).name
            for file_ in dataset.files
        }

    result = runner.invoke(
        cli, [
            'dataset', 'unlink', 'my-dataset', '--include', new_file.basename,
            '-y'
        ]
    )
    assert 0 == result.exit_code

    with client.with_dataset('my-dataset') as dataset:
        assert new_file.basename not in [
            file_.path.name for file_ in dataset.files
        ]


def test_dataset_rm(tmpdir, runner, project, client):
    """Test removal of a dataset."""
    # try to delete non existing dataset
    result = runner.invoke(cli, ['dataset', 'rm'])
    assert 2 == result.exit_code

    result = runner.invoke(cli, ['dataset', 'rm', 'does-not-exist'])
    assert 2 == result.exit_code

    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    paths = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset'] + paths,
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    # try to delete a non empty dataset
    result = runner.invoke(cli, ['dataset', 'rm', 'my-dataset'])
    assert 0 == result.exit_code

    # check output
    assert 'OK' in result.output
    assert not client.load_dataset('my-dataset')

    result = runner.invoke(cli, ['doctor'], catch_exceptions=False)
    assert 0 == result.exit_code


def test_dataset_rm_commit(tmpdir, runner, project, client):
    """Test removal of a dataset repository state."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # try to delete a non empty dataset
    result = runner.invoke(cli, ['dataset', 'rm', 'my-dataset'])
    assert 0 == result.exit_code

    # check output
    assert 'OK' in result.output
    assert not client.load_dataset('my-dataset')

    # Dirty repository check.
    result = runner.invoke(cli, ['status'])
    assert 0 == result.exit_code


def test_dataset_overwrite_no_confirm(runner, project):
    """Check dataset overwrite behaviour without confirmation."""
    result = runner.invoke(cli, ['dataset', 'create', 'rokstar'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    result = runner.invoke(cli, ['dataset', 'create', 'rokstar'])
    assert 1 == result.exit_code
    assert 'OK' not in result.output


@pytest.mark.parametrize('dirty', [False, True])
def test_dataset_edit(runner, client, project, dirty):
    """Check dataset metadata editing."""
    if dirty:
        with (client.path / 'dirty_file').open('w') as fp:
            fp.write('a')

    result = runner.invoke(
        cli, ['dataset', 'create', 'dataset', '-t', 'original title']
    )
    assert 0 == result.exit_code

    creator1 = 'Forename1 Surname1 <name.1@mail.com> [Affiliation 1]'
    creator2 = 'Forename2 Surname2'

    result = runner.invoke(
        cli, [
            'dataset', 'edit', 'dataset', '-d', ' new description ', '-c',
            creator1, '-c', creator2
        ],
        catch_exceptions=False
    )
    assert 0 == result.exit_code
    assert 'Successfully updated: creators, description.' in result.output
    warning_msg = 'Warning: No email or wrong format for: Forename2 Surname2'
    assert warning_msg in result.output

    dataset = client.load_dataset('dataset')
    assert ' new description ' == dataset.description
    assert 'original title' == dataset.name
    assert {creator1, creator2} == {c.full_identity for c in dataset.creator}

    result = runner.invoke(
        cli, ['dataset', 'edit', 'dataset', '-t', ' new title '],
        catch_exceptions=False
    )
    assert 0 == result.exit_code
    assert 'Successfully updated: title.' in result.output

    dataset = client.load_dataset('dataset')
    assert ' new description ' == dataset.description
    assert 'new title' == dataset.name
    assert {creator1, creator2} == {c.full_identity for c in dataset.creator}


@pytest.mark.parametrize('dirty', [False, True])
def test_dataset_edit_no_change(runner, client, project, dirty):
    """Check metadata editing does not commit when there is no change."""
    if dirty:
        with (client.path / 'dirty_file').open('w') as fp:
            fp.write('a')

    result = runner.invoke(
        cli, ['dataset', 'create', 'dataset', '-t', 'original title']
    )
    assert 0 == result.exit_code

    commit_sha_before = client.repo.head.object.hexsha

    result = runner.invoke(
        cli, ['dataset', 'edit', 'dataset'], catch_exceptions=False
    )
    assert 0 == result.exit_code
    assert 'Nothing to update.' in result.output

    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_after == commit_sha_before


def test_dataset_date_created_format(runner, client, project):
    """Check format of date created field."""
    # Create a dataset.
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    path = client.get_dataset_path('dataset')
    assert path.exists()

    with path.open(mode='r') as fp:
        import dateutil.parser as dp
        data_yaml = yaml.safe_load(fp)

        assert 'created' in data_yaml
        assert dp.parse(data_yaml['created'])
        assert validate_iso8601(data_yaml['created'])


def test_dataset_file_date_created_format(tmpdir, runner, client, project):
    """Check format of date created field."""
    # Create a dataset.
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    path = client.get_dataset_path('dataset')
    assert path.exists()

    # Create data file.
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # Add data to dataset.
    result = runner.invoke(cli, ['dataset', 'add', 'dataset', str(new_file)])
    assert 0 == result.exit_code

    with path.open(mode='r') as fp:
        import dateutil.parser as dp
        data_yaml = yaml.safe_load(fp)

        assert 'created' in data_yaml
        assert 'files' in data_yaml
        assert dp.parse(data_yaml['files'][0]['added'])
        assert dp.parse(data_yaml['created'])
        assert validate_iso8601(data_yaml['created'])
        assert validate_iso8601(data_yaml['files'][0]['added'])


@pytest.mark.parametrize(
    'uri', [
        '10.5281/zenodo.3363060', 'doi:10.5281/zenodo.3363060',
        'https://zenodo.org/record/3363060'
    ]
)
def test_dataset_provider_resolution_zenodo(doi_responses, uri):
    """Check that zenodo uris resolve to ZenodoProvider."""
    provider, _ = ProviderFactory.from_uri(uri)
    assert type(provider) is ZenodoProvider


@pytest.mark.parametrize(
    'uri', [
        '10.7910/DVN/TJCLKP', 'doi:10.7910/DVN/TJCLKP',
        (
            'https://dataverse.harvard.edu/dataset.xhtml'
            '?persistentId=doi:10.7910/DVN/TJCLKP'
        )
    ]
)
def test_dataset_provider_resolution_dataverse(doi_responses, uri):
    """Check that dataverse URIs resolve to ``DataverseProvider``."""
    provider, _ = ProviderFactory.from_uri(uri)
    assert type(provider) is DataverseProvider


def test_dataset_tag(tmpdir, runner, project):
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    new_file = tmpdir.join('file')
    new_file.write(str('test'))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset',
         str(new_file)],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    # tag dataset
    result = runner.invoke(
        cli,
        ['dataset', 'tag', 'my-dataset', '1.0'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'tag', 'my-dataset', 'A', '-d', 'short descriptiön'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'tag', 'my-dataset', 'aBc9.34-11_55.t'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code


@pytest.mark.parametrize('form', ['tabular', 'json-ld'])
def test_dataset_ls_tags(tmpdir, runner, project, client, form):
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    new_file = tmpdir.join('file')
    new_file.write(str('test'))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset',
         str(new_file)],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    commit1 = client.repo.head.commit.hexsha

    # tag dataset
    result = runner.invoke(
        cli,
        ['dataset', 'tag', 'my-dataset', '1.0', '-d', 'first tag!'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    commit2 = client.repo.head.commit.hexsha

    result = runner.invoke(
        cli,
        ['dataset', 'tag', 'my-dataset', 'aBc9.34-11_55.t'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'ls-tags', 'my-dataset', '--format={}'.format(form)],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code
    assert '1.0' in result.output
    assert 'aBc9.34-11_55.t' in result.output
    assert 'first tag!' in result.output
    assert commit1 in result.output
    assert commit2 in result.output


def test_dataset_rm_tag(tmpdir, runner, project, client):
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    new_file = tmpdir.join('file')
    new_file.write(str('test'))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset',
         str(new_file)],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    commit1 = client.repo.head.commit.hexsha

    # tag dataset
    result = runner.invoke(
        cli,
        ['dataset', 'tag', 'my-dataset', '1.0', '-d', 'first tag!'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'ls-tags', 'my-dataset'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code
    assert '1.0' in result.output
    assert 'first tag!' in result.output
    assert commit1 in result.output

    result = runner.invoke(
        cli,
        ['dataset', 'rm-tags', 'my-dataset', '2.0'],
        catch_exceptions=False,
    )
    assert 2 == result.exit_code
    assert 'not found' in result.output

    result = runner.invoke(
        cli,
        ['dataset', 'rm-tags', 'my-dataset', '1.0'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'rm-tags', 'my-dataset', '1.0'],
        catch_exceptions=False,
    )
    assert 2 == result.exit_code
    assert 'not found' in result.output


def test_dataset_rm_tags_multiple(tmpdir, runner, project, client):
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    new_file = tmpdir.join('file')
    new_file.write(str('test'))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset',
         str(new_file)],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    for i in range(1, 4):
        # tag dataset
        result = runner.invoke(
            cli,
            ['dataset', 'tag', 'my-dataset',
             str(i)],
            catch_exceptions=False,
        )
        assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'rm-tags', 'my-dataset', '1', '2', '3'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code
    assert '1' not in result.output
    assert '2' not in result.output
    assert '3' not in result.output


def test_dataset_rm_tags_failure(tmpdir, runner, project, client):
    result = runner.invoke(
        cli,
        ['dataset', 'rm-tags', 'my-dataset', '1'],
        catch_exceptions=False,
    )

    assert 2 == result.exit_code
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create some data
    new_file = tmpdir.join('file')
    new_file.write(str('test'))

    # add data to dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset',
         str(new_file)],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'rm-tags', 'my-dataset', '1'],
        catch_exceptions=False,
    )
    assert 2 == result.exit_code


def test_dataset_clean_up_when_add_fails(runner, client):
    """Test project is cleaned when dataset add fails for a new dataset."""
    # add a non-existing path to a new dataset
    result = runner.invoke(
        cli,
        ['dataset', 'add', '--create', 'new-dataset', 'non-existing-file'],
        catch_exceptions=True,
    )

    assert 2 == result.exit_code
    ref = client.renku_path / 'refs' / 'datasets' / 'new-dataset'
    assert not ref.is_symlink() and not ref.exists()


def test_avoid_empty_commits(runner, client, directory_tree):
    """Test no empty commit is created when adding existing data."""
    runner.invoke(cli, ['dataset', 'create', 'my-dataset'])

    commit_sha_before = client.repo.head.object.hexsha
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset', directory_tree.strpath]
    )

    assert 0 == result.exit_code

    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_before != commit_sha_after

    commit_sha_before = commit_sha_after
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset', directory_tree.strpath]
    )
    assert 1 == result.exit_code

    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_before == commit_sha_after
    assert 'Error: File already exists in dataset.' in result.output


def test_multiple_dataset_commits(runner, client, directory_tree):
    """Check adding existing data to multiple datasets."""
    commit_sha_before = client.repo.head.object.hexsha
    result = runner.invoke(
        cli, ['dataset', 'add', '-c', 'my-dataset1', directory_tree.strpath]
    )

    assert 0 == result.exit_code

    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_before != commit_sha_after

    commit_sha_before = commit_sha_after
    result = runner.invoke(
        cli, ['dataset', 'add', '-c', 'my-dataset2', directory_tree.strpath]
    )
    assert 0 == result.exit_code

    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_before != commit_sha_after


def test_add_same_filename_multiple(runner, client, directory_tree):
    """Check adding same filename multiple times."""
    result = runner.invoke(
        cli, ['dataset', 'add', '-c', 'my-dataset1', directory_tree.strpath]
    )

    assert 0 == result.exit_code

    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset1', directory_tree.strpath]
    )
    assert 1 == result.exit_code
    assert 'Error: File already exists in dataset.' in result.output

    result = runner.invoke(
        cli,
        ['dataset', 'add', '--force', 'my-dataset1', directory_tree.strpath]
    )
    assert 1 == result.exit_code
    assert 'Error: There is nothing to commit.' in result.output

    result = runner.invoke(
        cli, [
            'dataset', 'add', '--force', 'my-dataset1', directory_tree.strpath,
            'README.md'
        ]
    )
    assert 0 == result.exit_code


@pytest.mark.parametrize('filename', ['.renku', '.renku/', 'Dockerfile'])
def test_add_protected_file(runner, client, filename):
    """Check adding a protected file."""
    result = runner.invoke(
        cli, ['dataset', 'add', '-c', 'my-dataset1', filename]
    )

    assert 1 == result.exit_code
    assert 'Error: The following paths are protected' in result.output


@pytest.mark.parametrize(
    'filename', ['.renkunotactuallyrenku', 'thisisnot.renku']
)
def test_add_nonprotected_file(runner, client, tmpdir, filename):
    """Check adding an 'almost' protected file."""

    new_file = tmpdir.join('filename')
    new_file.write(str('test'))

    result = runner.invoke(
        cli, ['dataset', 'add', '-c', 'my-dataset1',
              str(new_file)]
    )

    assert 0 == result.exit_code


def test_add_removes_local_path_information(runner, client, directory_tree):
    """Test local paths are removed when adding to a dataset."""
    result = runner.invoke(
        cli, ['dataset', 'add', '-c', 'my-dataset', directory_tree.strpath]
    )
    assert 0 == result.exit_code

    with client.with_dataset('my-dataset') as dataset:
        for file_ in dataset.files:
            assert file_.url.startswith('file://../')
            assert file_.url.endswith(file_.name)


def test_add_remove_credentials(runner, client, monkeypatch):
    """Check removal of credentials during adding of remote data files."""
    url = 'https://username:password@example.com/index.html'

    def get(u, *args, **kwargs):
        """Mocked response."""
        response = requests.Response()
        response._content = b'{}'
        response._content_consumed = True
        response.status_code = 200
        return response

    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code

    monkeypatch.setattr(requests, 'get', get)
    dataset = client.load_dataset('my-dataset')
    o = client._add_from_url(dataset, url, client.path, extract=False)

    assert 'https://example.com/index.html' == o[0]['url']


def test_pull_data_from_lfs(runner, client, tmpdir):
    """Test pulling data from LFS using relative paths."""
    data = tmpdir.join('data.txt')
    data.write('DATA')

    result = runner.invoke(cli, ['dataset', 'add', '-c', 'my-data', str(data)])
    assert 0 == result.exit_code

    relative_path = Path('data') / 'my-data' / 'data.txt'

    result = runner.invoke(cli, ['storage', 'pull', str(relative_path)])
    assert 0 == result.exit_code


def test_dataset_cmd_subdirectory(runner, project):
    """Check dataset command in sub directory."""
    # Ensure root.
    result = runner.invoke(cli, ['dataset'])
    assert 0 == result.exit_code

    # Ensure sub directory.
    with chdir(Path(project) / 'data'):
        result = runner.invoke(cli, ['dataset'])
        assert 0 == result.exit_code

    # Ensure a protected directory.
    with chdir(Path(project) / '.renku'):
        result = runner.invoke(cli, ['dataset'])
        assert 0 == result.exit_code


def test_add_external_files(runner, client, directory_tree):
    """Check adding external files."""
    result = runner.invoke(
        cli, [
            'dataset', 'add', '-c', '--external', 'my-data', '-d', 'files',
            directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code

    path = client.path / 'data' / 'my-data' / 'files' / 'file'
    assert path.exists()
    assert path.is_symlink()
    external_path = Path(directory_tree.strpath) / 'file'
    assert path.resolve() == external_path

    with client.with_dataset('my-data') as dataset:
        assert dataset.find_file('data/my-data/files/file') is not None


def test_add_external_file_multiple(runner, client, directory_tree):
    """Check adding external files multiple times."""
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', '--external', 'my-data',
            directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'add', '--external', 'my-data', directory_tree.strpath]
    )
    assert 1 == result.exit_code
    assert 'File already exists in dataset.' in result.output

    result = runner.invoke(
        cli, [
            'dataset', 'add', '--external', 'my-data', '--force',
            directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code


def test_overwrite_external_file(runner, client, directory_tree):
    """Check overwriting external and normal files."""
    # Add external file
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', '--external', 'my-data',
            directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code

    # Cannot add the same file
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-data', directory_tree.strpath]
    )
    assert 1 == result.exit_code
    assert 'File already exists in dataset.' in result.output

    # Can add the same file with --force
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-data', '--force', directory_tree.strpath]
    )
    assert 0 == result.exit_code
    assert [] == list(client.renku_pointers_path.rglob('*'))

    # Can add the same external file
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--external', 'my-data', '--force',
            directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code


def test_remove_external_file(runner, client, directory_tree):
    """Test removal of external files."""
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', '--external', 'my-data', '-d',
            'files', directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code

    targets_before = {
        str(p.resolve())
        for p in client.renku_pointers_path.rglob('*')
    }
    path = str(Path('data') / 'my-data' / 'files' / 'file')

    result = runner.invoke(cli, ['rm', str(path)])
    assert 0 == result.exit_code

    targets_after = {
        str(p.resolve())
        for p in client.renku_pointers_path.rglob('*')
    }

    removed = targets_before - targets_after
    assert 1 == len(removed)
    assert removed.pop().endswith('/file')


def test_unavailable_external_files(runner, client, directory_tree):
    """Check for external files that are not available."""
    result = runner.invoke(
        cli, [
            'dataset', 'add', '-c', '--external', 'my-data', '-d', 'files',
            directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code

    path = Path('data') / 'my-data' / 'files' / 'file'
    target = path.resolve()

    directory_tree.join('file').remove()
    assert not path.exists()

    # Update won't work
    result = runner.invoke(cli, ['dataset', 'update', '--external'])
    assert 2 == result.exit_code
    assert 'External file not found' in result.output

    # Renku doctor shows inaccessible files
    result = runner.invoke(cli, ['doctor'])
    assert 1 == result.exit_code
    assert 'There are missing external files.' in result.output
    assert str(path) in result.output
    assert str(target) in result.output


def test_external_and_link(runner, client, directory_tree):
    """Test cannot use --link and --external together."""
    result = runner.invoke(
        cli, [
            'dataset', 'add', '-c', '--external', '--link', 'my-data',
            directory_tree.strpath
        ]
    )
    assert 2 == result.exit_code
    assert 'Cannot use "--link" and "--external" together.' in result.output


def test_external_file_update(runner, client, directory_tree, project):
    """Check updating external files."""
    result = runner.invoke(
        cli, [
            'dataset', 'add', '-c', '--external', 'my-data', '-d', 'files',
            directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code

    directory_tree.join('file').write('some updates')

    path = str(Path('data') / 'my-data' / 'files' / 'file')
    previous_commit = client.find_previous_commit(path)

    result = runner.invoke(cli, ['dataset', 'update', '--external', 'my-data'])
    assert 0 == result.exit_code

    current_commit = client.find_previous_commit(path)
    assert current_commit != previous_commit


def test_workflow_with_external_file(
    runner, client, directory_tree, project, run
):
    """Check using external files in workflows."""
    result = runner.invoke(
        cli, [
            'dataset', 'add', '-c', '--external', 'my-data', '-d', 'files',
            directory_tree.strpath
        ]
    )
    assert 0 == result.exit_code

    source = Path(project) / 'data' / 'my-data' / 'files' / 'file'
    output = Path(project) / 'data' / 'output.txt'

    assert 0 == run(args=('run', 'wc', '-c'), stdin=source, stdout=output)

    previous_commit = client.find_previous_commit(output)

    # Update external file
    directory_tree.join('file').write('some updates')

    # Nothing is changed unless external files are updated
    result = runner.invoke(cli, ['status'])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['dataset', 'update', '--external', 'my-data'])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['status'])
    assert 1 == result.exit_code

    assert 0 == run(args=('update', ))

    result = runner.invoke(cli, ['status'])
    assert 0 == result.exit_code

    current_commit = client.find_previous_commit(source)
    assert current_commit != previous_commit



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service dataset jobs tests."""
import json

import pytest
from flaky import flaky
from git import Repo
from tests.service.views.test_dataset_views import assert_rpc_response

from renku.core.errors import DatasetExistsError, ParameterError
from renku.service.jobs.datasets import dataset_import
from renku.service.utils import make_project_path


@pytest.mark.parametrize(
    'url', [(
        'https://dev.renku.ch/projects/rokroskar/'
        'scratch-project/datasets/7eba3f50-1a19-4282-8a86-2497e0f43809/'
    )]
)
@pytest.mark.integration
@flaky(max_runs=30, min_passes=1)
def test_dataset_url_import_job(url, svc_client_with_repo):
    """Test dataset import via url."""
    svc_client, headers, project_id, url_components = svc_client_with_repo
    user = {'user_id': headers['Renku-User-Id']}
    payload = {
        'project_id': project_id,
        'dataset_uri': url,
    }
    response = svc_client.post(
        '/datasets.import',
        data=json.dumps(payload),
        headers=headers,
    )

    assert response
    assert_rpc_response(response)
    assert {'job_id', 'created_at'} == set(response.json['result'].keys())

    dest = make_project_path(
        user, {
            'owner': url_components.owner,
            'name': url_components.name
        }
    )

    old_commit = Repo(dest).head.commit
    job_id = response.json['result']['job_id']

    dataset_import(
        user,
        job_id,
        project_id,
        url,
    )

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha != new_commit.hexsha
    assert f'service: dataset import {url}' == new_commit.message

    response = svc_client.get(
        f'/jobs/{job_id}',
        headers=headers,
    )

    assert response
    assert_rpc_response(response)
    assert 'COMPLETED' == response.json['result']['state']


@pytest.mark.parametrize(
    'doi', [
        '10.5281/zenodo.3239980',
        '10.5281/zenodo.3188334',
        '10.7910/DVN/TJCLKP',
    ]
)
@pytest.mark.integration
@pytest.mark.service
@flaky(max_runs=30, min_passes=1)
def test_dataset_import_job(doi, svc_client_with_repo):
    """Test dataset import via doi."""
    svc_client, headers, project_id, url_components = svc_client_with_repo
    user = {'user_id': headers['Renku-User-Id']}
    payload = {
        'project_id': project_id,
        'dataset_uri': doi,
    }
    response = svc_client.post(
        '/datasets.import',
        data=json.dumps(payload),
        headers=headers,
    )

    assert response
    assert_rpc_response(response)
    assert {'job_id', 'created_at'} == set(response.json['result'].keys())

    dest = make_project_path(
        user, {
            'owner': url_components.owner,
            'name': url_components.name
        }
    )

    old_commit = Repo(dest).head.commit
    job_id = response.json['result']['job_id']

    dataset_import(
        user,
        job_id,
        project_id,
        doi,
    )

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha != new_commit.hexsha
    assert f'service: dataset import {doi}' == new_commit.message

    response = svc_client.get(
        f'/jobs/{job_id}',
        headers=headers,
    )
    assert response
    assert_rpc_response(response)
    assert 'COMPLETED' == response.json['result']['state']


@pytest.mark.parametrize(
    'doi,expected_err',
    [
        # not valid doi
        ('junkjunkjunk', 'Invalid parameter value'),
        # not existing doi
        ('10.5281/zenodo.11111111111111111', 'Invalid parameter value'),
    ]
)
@pytest.mark.integration
@pytest.mark.service
def test_dataset_import_junk_job(doi, expected_err, svc_client_with_repo):
    """Test dataset import."""
    svc_client, headers, project_id, url_components = svc_client_with_repo
    user = {'user_id': headers['Renku-User-Id']}
    payload = {
        'project_id': project_id,
        'dataset_uri': doi,
    }
    response = svc_client.post(
        '/datasets.import',
        data=json.dumps(payload),
        headers=headers,
    )

    assert response
    assert_rpc_response(response)
    assert {'job_id', 'created_at'} == set(response.json['result'].keys())

    dest = make_project_path(
        user, {
            'owner': url_components.owner,
            'name': url_components.name
        }
    )

    old_commit = Repo(dest).head.commit
    job_id = response.json['result']['job_id']

    with pytest.raises(ParameterError):
        dataset_import(
            user,
            job_id,
            project_id,
            doi,
        )

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha == new_commit.hexsha

    response = svc_client.get(
        f'/jobs/{job_id}',
        data=json.dumps(payload),
        headers=headers,
    )

    assert_rpc_response(response)
    extras = response.json['result']['extras']

    assert 'error' in extras
    assert expected_err in extras['error']


@pytest.mark.parametrize('doi', [
    '10.5281/zenodo.3634052',
])
@pytest.mark.integration
@pytest.mark.service
def test_dataset_import_twice_job(doi, svc_client_with_repo):
    """Test dataset import."""
    svc_client, headers, project_id, url_components = svc_client_with_repo
    user = {'user_id': headers['Renku-User-Id']}
    payload = {
        'project_id': project_id,
        'dataset_uri': doi,
    }
    response = svc_client.post(
        '/datasets.import',
        data=json.dumps(payload),
        headers=headers,
    )

    assert response
    assert_rpc_response(response)
    assert {'job_id', 'created_at'} == set(response.json['result'].keys())

    dest = make_project_path(
        user, {
            'owner': url_components.owner,
            'name': url_components.name
        }
    )

    old_commit = Repo(dest).head.commit
    job_id = response.json['result']['job_id']

    dataset_import(
        user,
        job_id,
        project_id,
        doi,
    )

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha != new_commit.hexsha

    with pytest.raises(DatasetExistsError):
        dataset_import(
            user,
            job_id,
            project_id,
            doi,
        )

    new_commit2 = Repo(dest).head.commit
    assert new_commit.hexsha == new_commit2.hexsha

    response = svc_client.get(
        f'/jobs/{job_id}',
        data=json.dumps(payload),
        headers=headers,
    )

    assert_rpc_response(response)
    extras = response.json['result']['extras']

    assert 'error' in extras
    assert 'Dataset exists' in extras['error']



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service entry point."""
import logging
import os
import uuid

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from flask import Flask
from flask_apispec import FlaskApiSpec
from flask_swagger_ui import get_swaggerui_blueprint

from renku.service.cache import cache
from renku.service.config import API_SPEC_URL, API_VERSION, CACHE_DIR, \
    OPENAPI_VERSION, SERVICE_NAME, SWAGGER_URL
from renku.service.utils.json_encoder import SvcJSONEncoder
from renku.service.views.cache import CACHE_BLUEPRINT_TAG, cache_blueprint, \
    list_projects_view, list_uploaded_files_view, project_clone, \
    upload_file_view
from renku.service.views.datasets import DATASET_BLUEPRINT_TAG, \
    add_file_to_dataset_view, create_dataset_view, dataset_blueprint, \
    list_dataset_files_view, list_datasets_view
from renku.service.views.jobs import JOBS_BLUEPRINT_TAG, jobs_blueprint, \
    list_jobs

logging.basicConfig(level=os.getenv('SERVICE_LOG_LEVEL', 'WARNING'))


def create_app():
    """Creates a Flask app with a necessary configuration."""
    app = Flask(__name__)
    app.secret_key = os.getenv('RENKU_SVC_SERVICE_KEY', uuid.uuid4().hex)
    app.json_encoder = SvcJSONEncoder
    app.config['UPLOAD_FOLDER'] = CACHE_DIR

    max_content_size = os.getenv('MAX_CONTENT_LENGTH')
    if max_content_size:
        app.config['MAX_CONTENT_LENGTH'] = max_content_size

    app.config['cache'] = cache

    build_routes(app)

    @app.route('/health')
    def health():
        import renku
        return 'renku repository service version {}\n'.format(
            renku.__version__
        )

    return app


def build_routes(app):
    """Register routes to given app instance."""
    app.config.update({
        'APISPEC_SPEC':
            APISpec(
                title=SERVICE_NAME,
                openapi_version=OPENAPI_VERSION,
                version=API_VERSION,
                plugins=[MarshmallowPlugin()],
            ),
        'APISPEC_SWAGGER_URL': API_SPEC_URL,
    })
    app.register_blueprint(cache_blueprint)
    app.register_blueprint(dataset_blueprint)
    app.register_blueprint(jobs_blueprint)

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL, API_SPEC_URL, config={'app_name': 'Renku Service'}
    )
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    docs = FlaskApiSpec(app)

    docs.register(upload_file_view, blueprint=CACHE_BLUEPRINT_TAG)
    docs.register(list_uploaded_files_view, blueprint=CACHE_BLUEPRINT_TAG)
    docs.register(project_clone, blueprint=CACHE_BLUEPRINT_TAG)
    docs.register(list_projects_view, blueprint=CACHE_BLUEPRINT_TAG)

    docs.register(create_dataset_view, blueprint=DATASET_BLUEPRINT_TAG)
    docs.register(add_file_to_dataset_view, blueprint=DATASET_BLUEPRINT_TAG)
    docs.register(list_datasets_view, blueprint=DATASET_BLUEPRINT_TAG)
    docs.register(list_dataset_files_view, blueprint=DATASET_BLUEPRINT_TAG)

    docs.register(list_jobs, blueprint=JOBS_BLUEPRINT_TAG)


app = create_app()

if __name__ == '__main__':
    app.run()



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project initialization tests."""

import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import pkg_resources
import pytest
from tests.utils import raises

from renku.core import errors
from renku.core.commands.init import TEMPLATE_MANIFEST, create_from_template, \
    fetch_template, read_template_manifest, validate_template
from renku.core.management.config import RENKU_HOME

TEMPLATE_URL = (
    'https://github.com/SwissDataScienceCenter/renku-project-template'
)
TEMPLATE_NAME = 'Basic Python Project'
TEMPLATE_REF = '0.1.5'
METADATA = {'name': 'myname', 'description': 'nodesc'}
FAKE = 'NON_EXISTING'

template_local = Path(pkg_resources.resource_filename('renku', 'templates'))


@pytest.mark.parametrize(
    'url, ref, result, error', [(TEMPLATE_URL, TEMPLATE_REF, True, None),
                                (FAKE, TEMPLATE_REF, None, errors.GitError),
                                (TEMPLATE_URL, FAKE, None, errors.GitError)]
)
@pytest.mark.integration
def test_fetch_template(url, ref, result, error):
    """Test fetching a template.

    It fetches a template from remote and verifies that the manifest
    file is there.
    """
    with TemporaryDirectory() as tempdir:
        with raises(error):
            manifest_file = fetch_template(url, ref, Path(tempdir))
            assert Path(tempdir) / TEMPLATE_MANIFEST == manifest_file
            assert manifest_file.exists()


def test_read_template_manifest():
    """Test reading template manifest file.

    It creates a fake manifest file and it verifies it's read propery.
    """
    with TemporaryDirectory() as tempdir:
        template_file = Path(tempdir) / TEMPLATE_MANIFEST

        # error on missing template file
        with raises(errors.InvalidTemplateError):
            manifest = read_template_manifest(Path(tempdir), checkout=False)

        template_file.touch(exist_ok=True)
        # error on invalid template file
        with raises(errors.InvalidTemplateError):
            manifest = read_template_manifest(Path(tempdir), checkout=False)

        template_file.write_text(
            '-\n'
            '  folder: first\n'
            '  name: Basic Project 1\n'
            '  description: Description 1\n'
            '-\n'
            '  folder: second\n'
            '  name: Basic Project 2\n'
            '  description: Description 2\n'
        )

        manifest = read_template_manifest(Path(tempdir), checkout=False)
        assert 2 == len(manifest)
        assert 'first' == manifest[0]['folder']
        assert 'second' == manifest[1]['folder']
        assert 'Basic Project 1' == manifest[0]['name']
        assert 'Description 2' == manifest[1]['description']

        template_file.write_text(
            '-\n'
            '  folder: first\n'
            '  description: Description 1\n'
        )
        with raises(errors.InvalidTemplateError):
            manifest = read_template_manifest(Path(tempdir), checkout=False)

        template_file.write_text(
            '-\n'
            '  name: Basic Project 2\n'
            '  description: Description 2\n'
        )
        with raises(errors.InvalidTemplateError):
            manifest = read_template_manifest(Path(tempdir), checkout=False)


@pytest.mark.integration
def test_fetch_template_and_read_manifest():
    """Test template fetch and manifest reading.

    It fetches a local template, reads the manifest, checkouts the
    template folders and verify they exist.
    """
    with TemporaryDirectory() as tempdir:
        template_path = Path(tempdir)
        fetch_template(TEMPLATE_URL, TEMPLATE_REF, template_path)
        manifest = read_template_manifest(template_path, checkout=True)
        for template in manifest:
            template_folder = template_path / template['folder']
            assert template_folder.exists()


def test_validate_template():
    """Test template validation.

    It validates each local template.
    """
    with TemporaryDirectory() as tempdir:
        temppath = Path(tempdir)
        # file error
        with raises(errors.InvalidTemplateError):
            validate_template(temppath)

        # folder error
        shutil.rmtree(str(tempdir))
        renku_dir = temppath / RENKU_HOME
        renku_dir.mkdir(parents=True)
        with raises(errors.InvalidTemplateError):
            validate_template(temppath)

        # valid template
        shutil.rmtree(str(tempdir))
        shutil.copytree(str(template_local), str(tempdir))
        manifest = read_template_manifest(Path(tempdir))
        for template in manifest:
            template_folder = temppath / template['folder']
            assert validate_template(template_folder) is True


def test_create_from_template(local_client):
    """Test repository creation from a template.

    It creates a renku projects from one of the local templates and it verifies
    the data are properly copied to the new renku project folder.
    """
    with TemporaryDirectory() as tempdir:
        temppath = Path(tempdir) / 'local'
        shutil.copytree(str(template_local), str(temppath))
        manifest = read_template_manifest(temppath)
        template_path = temppath / manifest[0]['folder']
        create_from_template(
            template_path, local_client, METADATA['name'],
            METADATA['description']
        )
        template_files = [
            f
            for f in local_client.path.glob('**/*') if '.git' not in str(f) and
            not str(f).endswith('.renku/metadata.yml')
        ]
        for template_file in template_files:
            expected_file = template_path / template_file.relative_to(
                local_client.path
            )
            assert expected_file.exists()



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache project related models."""
from walrus import DateTimeField, IntegerField, Model, TextField

from renku.service.cache.base import BaseCache
from renku.service.config import CACHE_PROJECTS_PATH


class Project(Model):
    """User project object."""

    __database__ = BaseCache.model_db

    created_at = DateTimeField()

    project_id = TextField(primary_key=True, index=True)
    user_id = TextField(index=True)

    clone_depth = IntegerField()
    git_url = TextField()

    name = TextField()
    fullname = TextField()
    email = TextField()
    owner = TextField()
    token = TextField()

    @property
    def abs_path(self):
        """Full path of cached project."""
        return CACHE_PROJECTS_PATH / self.user_id / self.owner / self.name



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph builder."""

import os
from collections import OrderedDict, defaultdict, deque
from pathlib import Path

import attr

from renku.core import errors
from renku.core.models.cwl import CommandLineTool
from renku.core.models.cwl.parameter import InputParameter, \
    WorkflowOutputParameter
from renku.core.models.cwl.types import PATH_TYPES
from renku.core.models.cwl.workflow import Workflow
from renku.core.models.entities import Collection, Entity
from renku.core.models.git import Range
from renku.core.models.provenance.activities import Activity, ProcessRun, Usage
from renku.core.models.provenance.processes import Process
from renku.core.models.provenance.qualified import Generation

LINK_CWL = CommandLineTool(
    baseCommand=['true'],
    requirements=[
        {
            'class': 'InlineJavascriptRequirement',
        },
        {
            'class': 'InitialWorkDirRequirement',
            'listing': '$(inputs.input_directory.listing)',
        },
    ],
    inputs={
        'input_directory': 'Directory',
        'filename': 'string',
    },
    outputs={
        'output_file': {
            'type': 'File',
            'outputBinding': {
                'glob': '$(inputs.filename)',
                # .slice(inputs.input_directory.basename.length)
            },
        },
    },
)


def _safe_path(filepath, can_be_cwl=False):
    """Check if the path should be used in output."""
    if isinstance(filepath, Path):
        filepath = str(filepath)

    # Should not be in ignore paths.
    if filepath in {'.gitignore', '.gitattributes'}:
        return False

    # Ignore everything in .renku ...
    if filepath.startswith('.renku'):
        # ... unless it can be a CWL.
        if can_be_cwl and filepath.endswith('.cwl'):
            return True
        return False

    return True


@attr.s(cmp=False)
class Graph(object):
    """Represent the provenance graph."""

    client = attr.ib()
    activities = attr.ib(default=attr.Factory(dict))
    generated = attr.ib(default=attr.Factory(dict))

    _sorted_commits = attr.ib(default=attr.Factory(list))
    _latest_commits = attr.ib(default=attr.Factory(dict))
    _nodes = attr.ib()
    _need_update = attr.ib(default=attr.Factory(dict))

    cwl_prefix = attr.ib(init=False)

    def __attrs_post_init__(self):
        """Derive basic informations."""
        self.cwl_prefix = self.client.cwl_prefix

    @_nodes.default
    def default_nodes(self):
        """Build node index."""
        self.generated = {}
        nodes = OrderedDict()

        for commit in reversed(self._sorted_commits):
            try:
                activity = self.activities[commit]

                # for node in reversed(list(activity.nodes)):
                #     key = (node.commit, node.path)
                #     if key in nodes:
                #         del nodes[key]
                #     nodes[key] = node

                nodes.update(((node.commit, node.path), node)
                             for node in reversed(list(activity.nodes)))

                if isinstance(activity, ProcessRun):
                    self.generated.update({
                        generation.entity._id: generation
                        for generation in activity.generated
                    })

            except KeyError:
                pass

        return nodes

    def need_update(self, node):
        """Return out-dated nodes."""
        if node is None:
            return

        skip = True
        if isinstance(node, ProcessRun):
            node = node.association.plan
            skip = False

        if node._id in self._need_update:
            return self._need_update[node._id]

        latest = self.latest(node)
        if latest:
            return self._need_update.setdefault(node._id, [node])

        need_update_ = []

        for parent in self.parents(node):
            # Skip Collections if it is not an input
            if skip and isinstance(parent, Collection):
                continue

            parent_updates = self.need_update(parent)
            if parent_updates:
                need_update_.extend(parent_updates)

        return self._need_update.setdefault(node._id, need_update_)

    def parents(self, node):
        """Return parents for a given node."""
        import warnings

        def _from_entity(entity, check_parents=True):
            """Find parent from entity."""
            try:
                return [self.generated[entity._id].activity]
            except KeyError:
                id_ = Path(entity._id)
                while check_parents and id_ != id_.parent:
                    try:
                        # TODO include selection step here
                        return [self.generated[str(id_)]]
                    except KeyError:
                        id_ = id_.parent
                return []

        if isinstance(node, Generation):
            result = [node.parent] if node.parent is not None else []
            if node.activity and isinstance(node.activity, ProcessRun):
                return result + [node.activity.association.plan]
            return result
        elif isinstance(node, Usage):
            return _from_entity(node.entity)
        elif isinstance(node, Entity):
            # Link files and directories and generations.
            return ([node.parent] if node.parent is not None else
                    []) + _from_entity(node, check_parents=False)
        elif isinstance(node, Process):
            # warnings.warn('Called on run {0}'.format(node), stacklevel=2)
            return self.parents(node.activity)
        elif isinstance(node, ProcessRun):
            return node.qualified_usage
        elif isinstance(node, Activity):
            warnings.warn('Called parents on {0}'.format(node), stacklevel=2)
            return []

        raise NotImplementedError(node)

    def latest(self, node):
        """Return a latest commit where the node was modified."""
        if node.path and node.path not in self._latest_commits:
            try:
                latest = Usage.from_revision(
                    node.client,
                    path=node.path,
                    # TODO support range queries
                    # revision='{0}'.format(node.commit.hexsha),
                ).commit
            except KeyError:
                latest = None

            self._latest_commits[node.path] = latest
        else:
            latest = self._latest_commits.get(node.path)

        if latest and latest != node.commit:
            return latest

    @property
    def nodes(self):
        """Return topologically sorted nodes."""
        return reversed(self._nodes.values())

    def normalize_path(self, path):
        """Normalize path relative to the Git workdir."""
        start = self.client.path.resolve()
        try:
            p = Path(path).resolve()
            p.relative_to(self.client.path)
        except ValueError:  # External file
            path = os.path.abspath(path)
        else:
            path = p
        return os.path.relpath(str(path), start=str(start))

    def _format_path(self, path):
        """Return a relative path based on the client configuration."""
        return os.path.relpath(str(self.client.path / path))

    def dependencies(self, revision='HEAD', paths=None):
        """Return dependencies from a revision or paths."""
        result = []

        if paths:
            paths = (self.normalize_path(path) for path in paths)
        else:
            if revision == 'HEAD':
                index = self.client.repo.index
            else:
                from git import IndexFile
                index = IndexFile.from_tree(self.client.repo, revision)

            paths = (path for path, _ in index.entries.keys())

        for path in paths:
            try:
                result.append(
                    Usage.from_revision(
                        self.client,
                        path=path,
                        revision=revision,
                    )
                )
            except KeyError:
                continue

        return result

    def process_dependencies(self, dependencies, visited=None):
        """Process given dependencies."""
        for dependency in dependencies:
            # We can't simply reuse information from submodules
            if dependency.client != self.client:
                continue
            self._latest_commits[dependency.path] = dependency.commit

        visited = visited or set()
        queue = deque(dependencies)

        while queue:
            processing = queue.popleft()

            if processing.commit in visited:
                continue

            # Mark as visited:
            visited.add(processing.commit)

            activity = processing.client.process_commit(processing.commit)
            if activity is None:
                continue

            self.activities[activity.commit] = activity

            # Iterate over parents.
            if isinstance(activity, ProcessRun):
                for entity in activity.qualified_usage:
                    for member in entity.entities:
                        if member.commit not in visited:
                            queue.append(member)

        from renku.core.models.sort import topological
        self._sorted_commits = topological({
            commit: activity.parents
            for commit, activity in self.activities.items()
        })
        self._nodes = self.default_nodes()

    def build(
        self, revision='HEAD', paths=None, dependencies=None, can_be_cwl=False
    ):
        """Build graph from paths and/or revision."""
        interval = Range.rev_parse(self.client.repo, revision)

        if dependencies is None:
            dependencies = self.dependencies(revision=revision, paths=paths)

        ignore = {
            commit
            for commit in self.client.repo.iter_commits(interval.start)
        } if interval.start else set()

        self.process_dependencies(dependencies, visited=ignore)

        return {
            self._nodes.get((dependency.commit, dependency.path), dependency)
            for dependency in dependencies
            if _safe_path(dependency.path, can_be_cwl=can_be_cwl)
        }

    @property
    def output_paths(self):
        """Return all output paths."""
        paths = set()
        for activity in self.activities.values():
            if isinstance(activity, ProcessRun):
                paths |= {path for path in activity.outputs.keys() if path}
        return paths

    def build_status(self, revision='HEAD', can_be_cwl=False):
        """Return files from the revision grouped by their status."""
        status = {
            'up-to-date': {},
            'outdated': {},
            'multiple-versions': {},
            'deleted': {},
        }

        dependencies = self.dependencies(revision=revision)
        current_files = self.build(
            dependencies=dependencies,
            can_be_cwl=can_be_cwl,
        )

        # TODO check only outputs
        paths = {}
        for commit in reversed(self._sorted_commits):
            activity = self.activities.get(commit)

            if isinstance(activity, ProcessRun):
                nodes = activity.nodes if can_be_cwl else activity.generated

                for node in nodes:
                    paths[node.path] = node

        # First find all up-to-date nodes.
        for node in paths.values():
            # for node in current_files:
            need_update = [
                dependency for dependency in self.need_update(node)
                if dependency.path != node.path
            ]

            if need_update:
                status['outdated'][node.path] = need_update
            else:
                status['up-to-date'][node.path] = node.commit

        # Merge all versions of used inputs in outdated file.
        multiple_versions = defaultdict(set)

        for need_update in status['outdated'].values():
            for node in need_update:
                multiple_versions[node.path].add(node)

        for node in current_files:
            if node.path in multiple_versions:
                multiple_versions[node.path].add(node)

        status['multiple-versions'] = {
            key: value
            for key, value in multiple_versions.items() if len(value) > 1
        }

        # Build a list of used files that have been deleted.
        current_paths = {node.path for node in current_files}
        status['deleted'] = {
            node.path: node
            for node in self.nodes
            if _safe_path(node.path, can_be_cwl=can_be_cwl) and
            node.path not in current_paths and
            not ((self.client.path / node.path).exists() or
                 (self.client.path / node.path).is_dir())
        }
        return status

    def siblings(self, node):
        """Return siblings for a given node.

        The key is part of the result set, hence to check if the node has
        siblings you should check the lenght is greater than 1.
        """
        parent = None

        if isinstance(node, Entity):
            if not node.parent:
                return {node}
            parent_siblings = self.siblings(node.parent) - {node.parent}
            return set(node.parent.members) | parent_siblings
        elif isinstance(node, Generation):
            parent = node.activity
        elif isinstance(node, Usage):
            parent = self.activities[node.commit]
        elif isinstance(node, Process):
            return {node}

        if parent is None or not isinstance(parent, ProcessRun):
            raise errors.InvalidOutputPath(
                'The file "{0}" was not created by a renku command. \n\n'
                'Check the file history using: git log --follow "{0}"'.format(
                    node.path
                )
            )

        return set(parent.generated)

    def ascwl(
        self,
        input_paths=None,
        output_paths=None,
        outputs=None,
        use_latest=True,
    ):
        """Serialize graph to CWL workflow.

        :param global_step_outputs: Make all step outputs global.
        """
        if output_paths is None:
            output_paths = {
                node.path
                for node in outputs if _safe_path(node.path)
            }

        workflow = Workflow()

        processes = set()
        stack = []

        output_keys = {(node.commit, node.path) for node in outputs}
        nodes = {(node.commit, node.path): node for node in self.nodes}

        def connect_file_to_directory(node):
            """Return step connecting file to a directory."""
            process = attr.evolve(
                LINK_CWL,
                inputs={
                    'input_directory': 'Directory',
                    'filename': {
                        'type': 'string',
                        'default':
                            str(Path(node.path).relative_to(node.parent.path)),
                    },
                }
            )
            process_run = ProcessRun(
                commit=node.commit,
                client=node.client,
                path=None,
                process=process,
                inputs={
                    node.parent.path:
                        Usage(
                            entity=node.parent,
                            role='input_directory',
                        ),
                },
                outputs={
                    node.path: 'output_file',
                },
            )

            for generated in process_run.generated:
                nodes[(generated.commit, generated.path)] = generated

            return process_run

        for node in self.nodes:
            if (node.commit, node.path) not in output_keys:
                continue

            process_run = None
            if isinstance(node, Entity) and not hasattr(node, 'activity'):
                process_run = connect_file_to_directory(node)

                stack.append(process_run)
                processes.add(process_run)

            else:
                assert hasattr(node, 'activity'), node
                assert isinstance(node.activity, ProcessRun)

                plan = node.activity.association.plan
                latest = self.latest(plan)
                if use_latest and latest:
                    plan = nodes[(latest, plan.path)]

                process_run = plan.activity

                if process_run not in processes:
                    stack.append(process_run)
                    processes.add(process_run)

        while stack:
            action = stack.pop()

            if not hasattr(action, 'inputs'):
                continue

            for path, dependency in action.inputs.items():
                # Do not follow defined input paths.
                if input_paths and path in input_paths:
                    continue

                node = nodes.get((dependency.commit, dependency.path),
                                 dependency)

                if isinstance(node, Generation):
                    process_run = node.activity
                elif isinstance(node, Collection) and node.parent:
                    raise NotImplementedError('Can not connect subdirectory')
                elif isinstance(node, Entity) and node.parent:
                    process_run = connect_file_to_directory(node)
                else:
                    process_run = None

                # Skip existing commits
                if process_run and isinstance(process_run, ProcessRun):
                    plan = process_run.association.plan
                    latest = self.latest(plan)
                    if process_run.path and use_latest and latest:
                        plan = nodes[(latest, plan.path)]

                    process_run = plan.activity

                    if process_run not in processes:
                        stack.append(process_run)
                        processes.add(process_run)

        steps = {
            tool: 'step_{0}'.format(tool_index)
            for tool_index, tool in enumerate(processes, 1)
        }

        def _source_name(commit, path):
            """Find source name for a node."""
            try:
                process_run = nodes[(commit, path)].activity
                output_id = process_run.outputs[path]
                return '{0}/{1}'.format(steps[process_run], output_id)
            except (KeyError, AttributeError):
                pass

        def _relative_default(client, default):
            """Evolve ``File`` or ``Directory`` path."""
            if isinstance(default, PATH_TYPES):
                path = os.path.abspath(client.workflow_path / default.path)
                return attr.evolve(default, path=path)
            return default

        input_index = 1

        for action, step_id in steps.items():
            tool = action.process

            ins = {}
            for path, dependency in action.inputs.items():
                alias = _source_name(dependency.commit, path)
                if alias:
                    ins[dependency.role] = alias

            outs = list(set(action.outputs.values()))

            for generated in action.generated:
                if generated.entity.path not in output_paths:
                    output_paths.add(generated.entity.path)
                    outputs.add(generated.entity)

            for input_ in tool.inputs:
                input_mapping = ins.get(input_.id)
                if input_mapping is None:
                    input_id = 'input_{0}'.format(input_index)
                    workflow.inputs.append(
                        InputParameter(
                            id=input_id,
                            type=input_.type,
                            default=_relative_default(
                                self.client, input_.default
                            ),
                        )
                    )
                    input_index += 1
                    ins[input_.id] = input_id

            workflow.add_step(
                run=self.client.path / action.path if action.path else tool,
                id=step_id,
                in_=ins,
                out=outs,
            )

        for index, node in enumerate(
            (node for node in outputs if node.path in output_paths)
        ):
            commit, path = node.commit, node.path
            id_ = 'output_{0}'.format(index)
            process_run = nodes[(commit, path)].activity

            if process_run.process is None or process_run.path is None:
                continue

            output_id = process_run.outputs[path]
            type_ = next(
                output for output in process_run.process.outputs
                if output.id == output_id
            ).type
            type_ = type_ if type_ == 'Directory' else 'File'
            output_source = _source_name(commit, path)

            if output_source is None:
                continue

            workflow.outputs.append(
                WorkflowOutputParameter(
                    id=id_,
                    type=type_,
                    outputSource=output_source,
                )
            )

        return workflow



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for dataset list files."""
from .tabulate import tabulate


def tabular(client, records, *, columns=None):
    """Format dataset files with a tabular output.

    :param client: LocalClient instance.
    :param records: Filtered collection.
    :param columns: List of columns to display
    """
    if not columns:
        columns = 'added,creators,dataset,full_path'

    return tabulate(
        collection=records,
        columns=columns,
        columns_mapping=DATASET_FILES_COLUMNS
    )


def jsonld(client, records, **kwargs):
    """Format dataset files as JSON-LD.

    :param client: LocalClient instance.
    :param records: Filtered collection.
    """
    from renku.core.models.json import dumps
    from renku.core.models.jsonld import asjsonld

    data = [asjsonld(record) for record in records]
    return dumps(data, indent=2)


DATASET_FILES_FORMATS = {
    'tabular': tabular,
    'json-ld': jsonld,
}
"""Valid formatting options."""

DATASET_FILES_COLUMNS = {
    'added': ('added', None),
    'creators': ('creators_csv', 'creators'),
    'creators_full': ('creators_full_csv', 'creators'),
    'dataset': ('title', 'dataset'),
    'full_path': ('full_path', None),
    'path': ('path', None),
    'short_name': ('short_name', 'dataset short_name'),
}



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache serializers."""
import time
import uuid
from datetime import datetime
from urllib.parse import urlparse

from marshmallow import Schema, ValidationError, fields, post_load, pre_load, \
    validates
from werkzeug.utils import secure_filename

from renku.core.errors import ConfigurationError
from renku.core.models.git import GitURL
from renku.service.config import PROJECT_CLONE_DEPTH_DEFAULT
from renku.service.serializers.rpc import JsonRPCResponse


def extract_file(request):
    """Extract file from Flask request.

    :raises: `ValidationError`
    """
    files = request.files
    if 'file' not in files:
        raise ValidationError('missing key: file')

    file = files['file']
    if file and not file.filename:
        raise ValidationError('wrong filename: {0}'.format(file.filename))

    if file:
        file.filename = secure_filename(file.filename)
        return file


class FileUploadRequest(Schema):
    """Request schema for file upload."""

    override_existing = fields.Boolean(missing=False)
    unpack_archive = fields.Boolean(missing=False)


class FileUploadContext(Schema):
    """Context schema for file upload."""

    created_at = fields.DateTime(missing=datetime.utcnow)
    file_id = fields.String(missing=lambda: uuid.uuid4().hex)

    content_type = fields.String(missing='unknown')
    file_name = fields.String(required=True)

    # measured in bytes (comes from stat() - st_size)
    file_size = fields.Integer(required=True)

    relative_path = fields.String(required=True)
    is_archive = fields.Boolean(missing=False)
    is_dir = fields.Boolean(required=True)
    unpack_archive = fields.Boolean(missing=False)


class FileUploadResponse(Schema):
    """Response schema for file upload."""

    files = fields.List(fields.Nested(FileUploadContext), required=True)


class FileUploadResponseRPC(JsonRPCResponse):
    """RPC response schema for file upload response."""

    result = fields.Nested(FileUploadResponse)


class FileListResponse(Schema):
    """Response schema for files listing."""

    files = fields.List(fields.Nested(FileUploadContext), required=True)


class FileListResponseRPC(JsonRPCResponse):
    """RPC response schema for files listing."""

    result = fields.Nested(FileListResponse)


class ProjectCloneRequest(Schema):
    """Request schema for project clone."""

    git_url = fields.String(required=True)
    depth = fields.Integer(missing=PROJECT_CLONE_DEPTH_DEFAULT)


class ProjectCloneContext(ProjectCloneRequest):
    """Context schema for project clone."""

    project_id = fields.String(missing=lambda: uuid.uuid4().hex)

    # measured in ms
    timestamp = fields.Integer(missing=time.time() * 1e+3)

    name = fields.String(required=True)
    fullname = fields.String(required=True)
    email = fields.String(required=True)
    owner = fields.String(required=True)
    token = fields.String(required=True)

    @validates('git_url')
    def validate_git_url(self, value):
        """Validates git url."""
        try:
            GitURL.parse(value)
        except ConfigurationError as e:
            raise ValidationError(str(e))

        return value

    @post_load()
    def format_url(self, data, **kwargs):
        """Format URL with a username and password."""
        git_url = urlparse(data['git_url'])

        url = 'oauth2:{0}@{1}'.format(data['token'], git_url.netloc)
        data['url_with_auth'] = git_url._replace(netloc=url).geturl()

        return data

    @pre_load()
    def set_owner_name(self, data, **kwargs):
        """Set owner and name fields."""
        git_url = GitURL.parse(data['git_url'])

        data['owner'] = git_url.owner
        data['name'] = git_url.name

        return data


class ProjectCloneResponse(Schema):
    """Response schema for project clone."""

    project_id = fields.String(required=True)
    git_url = fields.String(required=True)


class ProjectCloneResponseRPC(JsonRPCResponse):
    """RPC response schema for project clone response."""

    result = fields.Nested(ProjectCloneResponse)


class ProjectListResponse(Schema):
    """Response schema for project listing."""

    projects = fields.List(fields.Nested(ProjectCloneResponse), required=True)


class ProjectListResponseRPC(JsonRPCResponse):
    """RPC response schema for project listing."""

    result = fields.Nested(ProjectListResponse)



# -*- coding: utf-8 -*-
#
# Copyright 2017, 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Create an empty Renku project or reinitialize an existing one.

Starting a Renku project
~~~~~~~~~~~~~~~~~~~~~~~~

If you have an existing directory which you want to turn into a Renku project,
you can type:

.. code-block:: console

    $ cd ~/my_project
    $ renku init

or:

.. code-block:: console

    $ renku init ~/my_project

This creates a new subdirectory named ``.renku`` that contains all the
necessary files for managing the project configuration.

If provided directory does not exist, it will be created.

Updating an existing project
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

There are situations when the required structure of a Renku project needs
to be recreated or you have an **existing** Git repository. You can solve
these situation by simply adding the ``--force`` option.

.. code-block:: console

    $ git init .
    $ echo "# Example\nThis is a README." > README.md
    $ git add README.md
    $ git commit -m 'Example readme file'
    # renku init would fail because there is a git repository
    $ renku init --force

You can also enable the external storage system for output files, if it
was not installed previously.

.. code-block:: console

    $ renku init --force --external-storage

"""

import ast
import configparser
import os
from collections import OrderedDict, namedtuple
from pathlib import Path
from tempfile import mkdtemp

import attr
import click
import pkg_resources
from git import Repo

from renku.core import errors
from renku.core.commands.client import pass_local_client
from renku.core.commands.git import set_git_home
from renku.core.commands.init import create_from_template, fetch_template, \
    read_template_manifest
from renku.core.commands.options import option_use_external_storage
from renku.core.models.tabulate import tabulate

_GITLAB_CI = '.gitlab-ci.yml'
_DOCKERFILE = 'Dockerfile'
_REQUIREMENTS = 'requirements.txt'
CI_TEMPLATES = [_GITLAB_CI, _DOCKERFILE, _REQUIREMENTS]


def parse_variables(ctx, param, value):
    """Parse template variables to dictionary."""
    try:
        variables = ast.literal_eval(value)
    except ValueError:
        raise_template_error(value)
    if type(variables) is not dict:
        raise_template_error(value)
    return variables


def raise_template_error(value):
    """Raise template error with short explanation."""
    error_info = [
        '{0}'.format(value), 'Tip: a dictionary is expected',
        (
            'Example: --template-variables '
            '\'{ "variable_1": "string", "variable_2": 2 }\''
        )
    ]
    raise errors.ParameterError(
        '\n'.join(error_info), '"--template-variables"'
    )


def validate_name(ctx, param, value):
    """Validate a project name."""
    if not value:
        value = os.path.basename(ctx.params['path'].rstrip(os.path.sep))
    return value


def store_directory(value):
    """Store directory as a new Git home."""
    Path(value).mkdir(parents=True, exist_ok=True)
    set_git_home(value)
    return value


def create_template_sentence(templates, instructions=False):
    """Create templates choice sentence.

    :ref templates: list of templates coming from manifest file
    :ref instructions: add instructions
    """
    Template = namedtuple('Template', ['index', 'name', 'description'])
    templates_friendly = [
        Template(
            index=index + 1,
            name=template_elem['name'],
            description=template_elem['description'],
        ) for index, template_elem in enumerate(templates)
    ]

    text = tabulate(
        templates_friendly,
        headers=OrderedDict((
            ('index', 'Number'),
            ('name', 'Name'),
            ('description', 'Description'),
        ))
    )

    if not instructions:
        return text
    return '{0}\nPlease choose a template by typing the number'.format(text)


def is_path_empty(path):
    """Check if path contains files.

    :ref path: target path
    """
    gen = Path(path).glob('**/*')
    return not any(gen)


def check_git_user_config():
    """Check that git user information is configured."""
    dummy_git_folder = mkdtemp()
    repo = Repo.init(dummy_git_folder)
    git_config = repo.config_reader()
    try:
        git_config.get_value('user', 'name', None)
        git_config.get_value('user', 'email', None)
        return True
    except (configparser.NoOptionError, configparser.NoSectionError):
        return False


@click.command()
@click.argument(
    'path',
    default='.',
    type=click.Path(writable=True, file_okay=False, resolve_path=True),
)
@click.option(
    '--name',
    callback=validate_name,
    help='Provide a custom project name.',
)
@click.option('--template', help='Provide the name of the template to use.')
@click.option(
    '--template-source', help='Provide the templates repository url or path.'
)
@click.option(
    '--template-ref',
    default='master',
    help='Specify the reference to checkout on remote template repository.',
)
@click.option(
    '--template-variables',
    default='{}',
    callback=parse_variables,
    help=(
        'Provide custom values for template variables. It must be a python '
        'dictionary.\nExample: \'{ "variable_1": "string", "variable_2": 2 }\''
    )
)
@click.option('--description', help='Describe your project.')
@click.option(
    '--print-manifest', is_flag=True, help='Print templates manifest only.'
)
@click.option('--force', is_flag=True, help='Override target path.')
@option_use_external_storage
@pass_local_client
@click.pass_context
def init(
    ctx, client, use_external_storage, path, name, template, template_source,
    template_ref, template_variables, description, print_manifest, force
):
    """Initialize a project in PATH. Default is current path."""
    # verify dirty path
    if not is_path_empty(path) and not force and not print_manifest:
        raise errors.InvalidFileOperation(
            'Folder "{0}" is not empty. Please add --force '
            'flag to transform it into a Renku repository.'.format(str(path))
        )

    if not check_git_user_config():
        raise errors.ConfigurationError(
            'The user name and email are not configured. '
            'Please use the "git config" command to configure them.\n\n'
            '\tgit config --global --add user.name "John Doe"\n'
            '\tgit config --global --add user.email '
            '"john.doe@example.com"\n'
        )

    # select template source
    if template_source:
        click.echo(
            'Fetching template from {0}@{1}... '.format(
                template_source, template_ref
            ),
            nl=False
        )
        template_folder = Path(mkdtemp())
        fetch_template(template_source, template_ref, template_folder)
        template_manifest = read_template_manifest(
            template_folder, checkout=True
        )
        click.secho('OK', fg='green')
    else:
        template_folder = Path(
            pkg_resources.resource_filename('renku', 'templates')
        )
        template_manifest = read_template_manifest(template_folder)

    # select specific template
    repeat = False
    template_data = None
    if template:
        template_filtered = [
            template_elem for template_elem in template_manifest
            if template_elem['name'] == template
        ]
        if len(template_filtered) == 1:
            template_data = template_filtered[0]
        else:
            click.echo('The template "{0}" is not available.'.format(template))
            repeat = True

    if print_manifest:
        if template_data:
            click.echo(create_template_sentence([template_data]))
        else:
            click.echo(create_template_sentence(template_manifest))
        return

    if not template or repeat:
        templates = [template_elem for template_elem in template_manifest]
        if len(templates) == 1:
            template_data = templates[0]
        else:
            template_num = click.prompt(
                text=create_template_sentence(templates, True),
                type=click.IntRange(1, len(templates)),
                show_default=False,
                show_choices=False
            )
            template_data = templates[template_num - 1]

    # set local path and storage
    store_directory(path)
    if not client.use_external_storage:
        use_external_storage = False
    ctx.obj = client = attr.evolve(
        client, path=path, use_external_storage=use_external_storage
    )
    if not is_path_empty(path):
        from git import GitCommandError
        try:
            commit = client.find_previous_commit('*')
            branch_name = 'pre_renku_init_{0}'.format(commit.hexsha[:7])
            with client.worktree(
                path=path,
                branch_name=branch_name,
                commit=commit,
                merge_args=[
                    '--no-ff', '-s', 'recursive', '-X', 'ours',
                    '--allow-unrelated-histories'
                ]
            ):
                click.echo(
                    'Saving current data in branch {0}'.format(branch_name)
                )
        except AttributeError:
            click.echo('Warning! Overwriting non-empty folder.')
        except GitCommandError as e:
            click.UsageError(e)

    # clone the repo
    template_path = template_folder / template_data['folder']
    click.echo('Initializing new Renku repository... ', nl=False)
    with client.lock:
        try:
            create_from_template(
                template_path, client, name, description, template_variables,
                force
            )
        except FileExistsError as e:
            raise click.UsageError(e)

    # Install git hooks
    from .githooks import install
    ctx.invoke(install, force=force)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache file related models."""
from walrus import BooleanField, DateTimeField, IntegerField, Model, TextField

from renku.service.cache.base import BaseCache
from renku.service.config import CACHE_UPLOADS_PATH


class File(Model):
    """User file object."""

    __database__ = BaseCache.model_db

    created_at = DateTimeField()

    file_id = TextField(primary_key=True, index=True)
    user_id = TextField(index=True)

    content_type = TextField()
    file_name = TextField()

    file_size = IntegerField()

    relative_path = TextField()
    is_archive = BooleanField()
    is_dir = BooleanField()
    unpack_archive = BooleanField()

    @property
    def abs_path(self):
        """Full path of cached file."""
        return CACHE_UPLOADS_PATH / self.user_id / self.relative_path

    def valid_file(self):
        """Ensure a file exists."""
        if self.abs_path.exists():
            self.is_dir = self.abs_path.is_dir()
            return True

        return False



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Get and set Renku repository or global options.

Set values
~~~~~~~~~~

You can set various Renku configuration options, for example the image registry
URL, with a command like:

.. code-block:: console

    $ renku config registry https://registry.gitlab.com/demo/demo

By default, configuration is stored locally in the project's directory. Use
``--global`` option to store configuration for all projects in your home
directory.

Remove values
~~~~~~~~~~~~~

To remove a specific key from configuration use:

.. code-block:: console

    $ renku config --remove registry

By default, only local configuration is searched for removal. Use ``--global``
option to remove a global configuration value.

Query values
~~~~~~~~~~~~

You can display all configuration values with:

.. code-block:: console

    $ renku config

Both local and global configuration files are read. Values in local
configuration take precedence over global values. Use ``--local`` or
``--global`` flag to read corresponding configuration only.

You can provide a KEY to display only its value:

.. code-block:: console

    $ renku config registry
    https://registry.gitlab.com/demo/demo
"""
import click

from renku.core import errors
from renku.core.commands.config import read_config, update_config


@click.command()
@click.argument('key', required=False, default=None)
@click.argument('value', required=False, default=None)
@click.option('--remove', is_flag=True, help='Remove specified key.')
@click.option(
    '--local',
    'local_only',
    is_flag=True,
    help='Read/store from/to local configuration only.'
)
@click.option(
    '--global',
    'global_only',
    is_flag=True,
    help='Read/store from/to global configuration only.'
)
def config(key, value, remove, local_only, global_only):
    """Manage configuration options."""
    is_write = value is not None

    if is_write and remove:
        raise errors.UsageError('Cannot remove and set at the same time.')
    if remove and not key:
        raise errors.UsageError('KEY is missing.')
    if local_only and global_only:
        raise errors.UsageError('Cannot use --local and --global together.')

    if remove:
        update_config(key, remove=remove, global_only=global_only)
    elif is_write:
        update_config(key, value=value, global_only=global_only)
    else:
        value = read_config(key, local_only, global_only)
        click.secho(value)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2019- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLI tests."""
import sys
import tempfile

import pytest

from renku.cli import cli


@pytest.mark.parametrize(
    'cmd', [
        ['config'],
        ['dataset'],
        ['doctor'],
        ['githooks'],
        ['log'],
        ['migrate'],
        ['mv'],
        ['rerun'],
        ['rm'],
        ['run'],
        ['show'],
        ['status'],
        ['storage'],
        ['update'],
        ['workflow'],
    ]
)
def test_cli_initialization_err(cmd, runner):
    """Test correct exception raise within non-renku repository."""
    from renku.core.utils.contexts import chdir

    with tempfile.TemporaryDirectory() as tmpdir:
        with chdir(tmpdir):
            result = runner.invoke(cli, cmd)
            assert 2 == result.exit_code

            expected_output = (
                'Error: `.` is not a renku repository.\n'
                'To initialize this as a '
                'renku repository use: `renku init`\n'
            )
            assert expected_output == result.output


@pytest.mark.parametrize(
    'cmd',
    [
        # NOTE: Clone command covered through integration tests.
        ['config', '--help'],
        ['dataset', '--help'],
        ['doctor', '--help'],
        ['githooks', '--help'],
        ['log', '--help'],
        ['migrate', '--help'],
        ['mv', '--help'],
        ['rerun', '--help'],
        ['rm', '--help'],
        ['run', '--help'],
        ['show', '--help'],
        ['status', '--help'],
        ['storage', '--help'],
        ['update', '--help'],
        ['workflow', '--help'],
        ['init', '--template', 'Basic Python Project', '--force'],
        ['init', '--help'],
        ['help'],
        ['--help'],
    ]
)
def test_cli_initialization_no_err_help(cmd, runner):
    """Test allowed commands within non-renku repository."""
    from renku.core.utils.contexts import chdir
    sys.argv = cmd
    with tempfile.TemporaryDirectory() as tmpdir:
        with chdir(tmpdir):
            result = runner.invoke(cli, cmd)
            assert 0 == result.exit_code



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service worker."""
import os
from contextlib import contextmanager

from rq import Worker
from rq.logutils import setup_loghandlers

from renku.core.errors import ConfigurationError, UsageError
from renku.service.jobs.queues import QUEUES, WorkerQueues


@contextmanager
def worker(queue_list):
    """Creates worker object."""

    def build_worker():
        """Build worker."""
        setup_loghandlers(os.getenv('RQ_WORKER_LOG_LEVEL', 'WARNING'))
        rq_worker = Worker(queue_list, connection=WorkerQueues.connection)
        return rq_worker

    yield build_worker()


def check_queues(queue_list):
    """Check if listening on specified queues is possible."""
    for queue in queue_list:
        if queue not in QUEUES:
            err_msg = (
                'invalid queue name: {0}\n\n'
                'valid queue names: \n{1}'.format(queue, '\n'.join(QUEUES))
            )
            raise UsageError(err_msg)


def start_worker(queue_list):
    """Start worker."""
    check_queues(queue_list)
    with worker(queue_list) as rq_worker:
        rq_worker.work()


if __name__ == '__main__':
    queues = os.getenv('RENKU_SVC_WORKER_QUEUES')

    if not queues:
        raise ConfigurationError((
            'Worker queues not specified. '
            'Please, set RENKU_SVC_WORKER_QUEUES environment variable.'
        ))

    start_worker([
        queue_name.strip() for queue_name in queues.strip().split(',')
    ])



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cleanup jobs."""
import os
import shutil
import time
from datetime import datetime

from renku.service.cache import FileManagementCache, ProjectManagementCache


def cache_files_cleanup():
    """Cache files a cleanup job."""
    ttl = int(os.getenv('RENKU_SVC_CLEANUP_TTL_FILES', 1800))
    cache = FileManagementCache()

    for user, files in cache.user_files():
        for file in files:
            # If record does not contain created_at,
            # it means its an old record, and we should mark it for deletion.
            created_at = (file.created_at -
                          datetime.utcfromtimestamp(0)).total_seconds() * 1e+3

            if not file.created_at:
                created_at = (time.time() - ttl) * 1e+3

            old = ((time.time() * 1e+3) - created_at) / 1e+3

            if old >= ttl:
                file_path = file.abs_path
                if file_path.exists() and file_path.is_file():
                    try:
                        file_path.unlink()
                    except FileNotFoundError:
                        pass

                if file_path.exists() and file_path.is_dir():
                    shutil.rmtree(str(file_path))

                file.delete()


def cache_project_cleanup():
    """Cache project a cleanup job."""
    ttl = int(os.getenv('RENKU_SVC_CLEANUP_TTL_PROJECTS', 1800))
    cache = ProjectManagementCache()

    for user, projects in cache.user_projects():
        for project in projects:

            # If record does not contain created_at,
            # it means its an old record, and we should mark it for deletion.
            created_at = (project.created_at -
                          datetime.utcfromtimestamp(0)).total_seconds() * 1e+3

            if not project.created_at:
                created_at = (time.time() - ttl) * 1e+3

            old = ((time.time() * 1e+3) - created_at) / 1e+3

            if old >= ttl:
                project_path = project.abs_path
                if project_path.exists():
                    shutil.rmtree(str(project_path))

                project.delete()



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``init`` command."""
from pathlib import Path

import pytest
from tests.core.commands.test_init import TEMPLATE_NAME, TEMPLATE_REF, \
    TEMPLATE_URL

from renku.cli import cli
from renku.cli.init import create_template_sentence

INIT = ['init', 'test-new-project', '--template', TEMPLATE_NAME]
INIT_REMOTE = [
    '--template-source', TEMPLATE_URL, '--template-ref', TEMPLATE_REF
]
INIT_FORCE = ['--force']
INIT_VARIABLES = ['--template-variables']


def test_template_selection_helpers():
    templates = [{
        'name': 'Template Python',
        'description': 'Description Python'
    }, {
        'name': 'Template R',
        'description': 'Description R'
    }]
    instructions = 'Please choose a template by typing the number'
    sentence = create_template_sentence(templates)
    stripped_sentence = ' '.join(sentence.split())
    assert '1 Template Python Description Python' in stripped_sentence
    assert '2 Template R Description R' in stripped_sentence
    assert instructions not in stripped_sentence
    full_sentence = create_template_sentence(templates, True)
    assert instructions in full_sentence


@pytest.mark.integration
def test_init(isolated_runner):
    """Test project initialization from template."""
    # create the project
    new_project = Path('test-new-project')
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, INIT)
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / '.renku').exists()
    assert (new_project / '.renku' / 'renku.ini').exists()
    assert (new_project / '.renku' / 'metadata.yml').exists()

    # try to re-create in the same folder
    result_re = isolated_runner.invoke(cli, INIT)
    assert 0 != result_re.exit_code

    # force re-create in the same folder
    result_re = isolated_runner.invoke(cli, INIT + INIT_FORCE)
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / '.renku').exists()
    assert (new_project / '.renku' / 'renku.ini').exists()
    assert (new_project / '.renku' / 'metadata.yml').exists()


@pytest.mark.integration
def test_init_force_in_empty_dir(isolated_runner):
    """Run init --force in empty directory."""
    new_project = Path('test-new-project')
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, INIT + INIT_FORCE)
    assert 0 == result.exit_code


@pytest.mark.integration
def test_init_force_in_dirty_dir(isolated_runner):
    """Run init --force in dirty directory."""
    new_project = Path('test-new-project')
    assert not new_project.exists()

    new_project.mkdir(parents=True)
    random_file = new_project / 'random_file.txt'
    with random_file.open('w') as dest:
        dest.writelines(['random text'])
    assert random_file.exists()

    result = isolated_runner.invoke(cli, INIT + INIT_FORCE)
    assert random_file.exists()
    assert 0 == result.exit_code


@pytest.mark.integration
def test_init_on_cloned_repo(isolated_runner, data_repository):
    """Run init --force in directory containing another repo."""
    new_project = Path('test-new-project')
    import shutil
    shutil.copytree(str(data_repository.working_dir), str(new_project))
    assert new_project.exists()

    # try to create in a dirty folder
    result = isolated_runner.invoke(cli, INIT)
    assert 0 != result.exit_code

    # force re-create in the same folder
    result = isolated_runner.invoke(cli, INIT + INIT_FORCE)
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / '.renku').exists()
    assert (new_project / '.renku' / 'renku.ini').exists()
    assert (new_project / '.renku' / 'metadata.yml').exists()


@pytest.mark.integration
def test_init_remote(isolated_runner):
    """Test project initialization from a remote template."""
    # create the project
    new_project = Path('test-new-project')
    assert not new_project.exists()
    result = isolated_runner.invoke(cli, INIT + INIT_REMOTE)
    assert 0 == result.exit_code
    assert new_project.exists()
    assert (new_project / '.renku').exists()
    assert (new_project / '.renku' / 'renku.ini').exists()
    assert (new_project / '.renku' / 'metadata.yml').exists()


@pytest.mark.integration
def test_init_with_variables(isolated_runner):
    """Test project initialization using custom metadata."""
    # create the project
    new_project = Path('test-new-project')
    assert not new_project.exists()
    metadata = 'not_dictionary'
    result = isolated_runner.invoke(cli, INIT + INIT_VARIABLES + [metadata])
    assert 0 != result.exit_code
    metadata = '{"correct": "dictionary"}'
    result = isolated_runner.invoke(cli, INIT + INIT_VARIABLES + [metadata])
    assert 0 == result.exit_code



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache job related models."""
from datetime import datetime

from marshmallow import Schema, fields, post_load

from renku.service.cache.models.job import USER_JOB_STATE_ENQUEUED, Job


class JobSchema(Schema):
    """Job serialization."""

    created_at = fields.DateTime(missing=datetime.utcnow)
    updated_at = fields.DateTime(missing=datetime.utcnow)

    job_id = fields.String(required=True)
    user_id = fields.String(required=True)

    state = fields.String(required=False, missing=USER_JOB_STATE_ENQUEUED)
    extras = fields.Dict(required=False)

    @post_load
    def make_job(self, data, **options):
        """Construct job object."""
        return Job(**data)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert models to Common Workflow Language."""

import os
from pathlib import Path

import attr
from attr._compat import iteritems
from attr._funcs import has
from attr._make import fields
from pyld import jsonld

from renku.core.models.jsonld import asjsonld
from renku.core.models.locals import ReferenceMixin, with_reference


class CWLType(type):
    """Register CWL types."""

    def __init__(cls, name, bases, namespace):
        """Register CWL types."""
        super(CWLType, cls).__init__(name, bases, namespace)
        if not hasattr(cls, 'registry'):
            cls.registry = dict()
        if name in cls.registry:
            raise ValueError('Duplicate CWL class {0} definition'.format(name))
        cls.registry[name] = cls


def type_from_metadata(metadata_type, value):
    """Instantiates a metadata class from cwl metadata."""
    context = None

    if '@context' in value:
        context = {'@context': value['@context']}
    elif hasattr(metadata_type, '_jsonld_context'):
        context = {'@context': metadata_type._jsonld_context}

    if context:
        value = jsonld.compact(value, context)
    return metadata_type.from_jsonld(value)


class CWLClass(ReferenceMixin, metaclass=CWLType):
    """Include ``class`` field in serialized object."""

    @classmethod
    def from_cwl(cls, data, __reference__=None):
        """Return an instance from CWL data."""
        exclude_properties = ['class', '$namespaces', '@reverse']
        class_name = data.get('class', None)
        cls = cls.registry.get(class_name, cls)

        if '$namespaces' in data:
            # handle custom metadata
            keys = data.keys()

            metadata_keys = [(k, False) for k in keys if ':' in k]

            if '@reverse' in keys:
                metadata_keys.extend((k, True) for k in data['@reverse'].keys()
                                     if ':' in k)

            attrs = fields(cls)

            for a in attrs:
                # map custom metadata
                if 'cwl_metadata' not in a.metadata:
                    continue

                metadata = a.metadata['cwl_metadata']

                k = (metadata.get('property'), metadata.get('reverse', False))

                if k not in metadata_keys:
                    continue

                metadata_type = metadata.get('type')

                if not metadata_type:
                    raise ValueError('CWL metadata type not specified')

                if metadata.get('reverse', False):
                    metadata_value = data['@reverse'][metadata['property']]
                else:
                    metadata_value = data[metadata['property']]
                    exclude_properties.append(metadata['property'])

                if isinstance(metadata_value, list):
                    data[a.name] = [
                        type_from_metadata(metadata_type, v)
                        for v in metadata_value
                    ]
                else:
                    data[a.name
                         ] = type_from_metadata(metadata_type, metadata_value)

        if __reference__:
            with with_reference(__reference__):
                self = cls(
                    **{
                        k: v
                        for k, v in iteritems(data)
                        if k not in exclude_properties
                    }
                )
        else:
            self = cls(
                **{
                    k: v
                    for k, v in iteritems(data) if k not in exclude_properties
                }
            )
        return self

    @classmethod
    def from_yaml(cls, path):
        """Return an instance from a YAML file."""
        import yaml

        with path.open(mode='r') as fp:
            self = cls.from_cwl(yaml.safe_load(fp), __reference__=path)

        return self


def mapped(cls, key='id', **kwargs):
    """Create list of instances from a mapping."""
    kwargs.setdefault('metadata', {})
    kwargs['metadata']['jsonldPredicate'] = {'mapSubject': key}
    kwargs.setdefault('default', attr.Factory(list))

    def converter(value):
        """Convert mapping to a list of instances."""
        if isinstance(value, dict):
            result = []
            for k, v in iteritems(value):
                if not hasattr(cls, 'from_cwl'):
                    vv = dict(v)
                    vv[key] = k
                else:
                    vv = attr.evolve(cls.from_cwl(v), **{key: k})
                result.append(vv)
        else:
            result = value

        def fix_keys(data):
            """Fix names of keys."""
            for a in fields(cls):
                a_name = a.name.rstrip('_')
                if a_name in data:
                    yield a.name, data[a_name]

        return [
            cls(**{kk: vv
                   for kk, vv in fix_keys(v)}) if not isinstance(v, cls) else v
            for v in result
        ]

    kwargs['converter'] = converter
    return attr.ib(**kwargs)


def ascwl(
    inst,
    recurse=True,
    filter=None,
    dict_factory=dict,
    retain_collection_types=False,
    basedir=None,
):
    """Return the ``attrs`` attribute values of *inst* as a dict.

    Support ``jsonldPredicate`` in a field metadata for generating
    mappings from lists.

    Adapted from ``attr._funcs``.
    """
    attrs = fields(inst.__class__)
    rv = dict_factory()

    def convert_value(v):
        """Convert special types."""
        if isinstance(v, Path):
            v = str(v)
            return os.path.relpath(v, str(basedir)) if basedir else v
        return v

    def convert_cwl_metadata(rv, v, meta):
        """Convert custom metadata to cwl."""
        if not v:
            return

        if isinstance(v, (list, tuple)):
            result = [asjsonld(vv) for vv in v]
        else:
            result = asjsonld(v)

        if 'reverse' in meta and meta['reverse']:
            if '@reverse' not in rv:
                rv['@reverse'] = {}
            rv['@reverse'][meta['property']] = result
        else:
            rv[meta['property']] = result

        if '$namespaces' not in rv:
            rv['$namespaces'] = {}

        rv['$namespaces'][meta['prefix']] = meta['namespace']

    for a in attrs:
        if a.name.startswith('__'):
            continue

        a_name = a.name.rstrip('_')
        v = getattr(inst, a.name)
        if filter is not None and not filter(a, v):
            continue
        if 'cwl_metadata' in a.metadata:
            convert_cwl_metadata(rv, v, a.metadata['cwl_metadata'])
            continue
        if recurse is True:
            if has(v.__class__):
                rv[a_name] = ascwl(
                    v,
                    recurse=True,
                    filter=filter,
                    dict_factory=dict_factory,
                    basedir=basedir,
                )

            elif isinstance(v, (tuple, list, set)):
                cf = v.__class__ if retain_collection_types is True else list
                rv[a_name] = cf([
                    ascwl(
                        i,
                        recurse=True,
                        filter=filter,
                        dict_factory=dict_factory,
                        basedir=basedir,
                    ) if has(i.__class__) else i for i in v
                ])

                if 'jsonldPredicate' in a.metadata:
                    k = a.metadata['jsonldPredicate'].get('mapSubject')
                    if k:
                        vv = dict_factory()
                        for i in rv[a_name]:
                            kk = i.pop(k)
                            vv[kk] = i
                        rv[a_name] = vv

            elif isinstance(v, dict):
                df = dict_factory
                rv[a_name] = df((
                    ascwl(
                        kk,
                        dict_factory=df,
                        basedir=basedir,
                    ) if has(kk.__class__) else convert_value(kk),
                    ascwl(
                        vv,
                        dict_factory=df,
                        basedir=basedir,
                    ) if has(vv.__class__) else vv
                ) for kk, vv in iteritems(v))
            else:
                rv[a_name] = convert_value(v)
        else:
            rv[a_name] = convert_value(v)

    if isinstance(inst, CWLClass):
        rv['class'] = inst.__class__.__name__

    return rv
